import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
from sklearn.model_selection import KFold, StratifiedKFold

from bofire.data_models.enum import OutputFilteringEnum
from bofire.data_models.features.api import (
    CategoricalInput,
    CategoricalOutput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.surrogates.diagnostics import CvResult, CvResults
from bofire.surrogates.surrogate import Surrogate


class TrainableSurrogate(ABC):
    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL

    def fit(self, experiments: pd.DataFrame, options: Optional[Dict] = None):
        # validate
        experiments = self.inputs.validate_experiments(  # type: ignore
            experiments, strict=False
        )
        experiments = self.outputs.validate_experiments(experiments)  # type: ignore
        # preprocess
        experiments = self._preprocess_experiments(experiments)
        X = experiments[self.inputs.get_keys()]  # type: ignore
        # TODO: output feature validation
        Y = experiments[self.outputs.get_keys()]  # type: ignore
        # fit
        options = options or {}
        self._fit(X=X, Y=Y, **options)  # type: ignore

    def _preprocess_experiments(self, experiments: pd.DataFrame) -> pd.DataFrame:
        if self._output_filtering is None:
            return experiments
        elif self._output_filtering == OutputFilteringEnum.ALL:
            return self.outputs.preprocess_experiments_all_valid_outputs(  # type: ignore
                experiments=experiments,
                output_feature_keys=self.outputs.get_keys(),  # type: ignore
            )
        elif self._output_filtering == OutputFilteringEnum.ANY:
            return self.outputs.preprocess_experiments_any_valid_outputs(  # type: ignore
                experiments=experiments,
                output_feature_keys=self.outputs.get_keys(),  # type: ignore
            )
        else:
            raise ValueError("Unknown output filtering option requested.")

    @abstractmethod
    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame, **kwargs):
        pass

    def cross_validate(
        self,
        experiments: pd.DataFrame,
        folds: int = -1,
        include_X: bool = False,
        include_labcodes: bool = False,
        random_state: Optional[int] = None,
        stratified_feature: Optional[str] = None,
        hooks: Optional[
            Dict[
                str,
                Callable[
                    [
                        Surrogate,
                        pd.DataFrame,
                        pd.DataFrame,
                        pd.DataFrame,
                        pd.DataFrame,
                    ],
                    Any,
                ],
            ]
        ] = None,
        hook_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> Tuple[CvResults, CvResults, Dict[str, List[Any]]]:
        """Perform a cross validation for the provided training data.

        Args:
            experiments (pd.DataFrame): Data on which the cross validation should be performed.
            folds (int, optional): Number of folds. -1 is equal to LOO CV. Defaults to -1.
            include_X (bool, optional): If true the X values of the fold are written to respective CvResult objects for
                later analysis. Defaults to False.
            random_state (int, optional): Controls the randomness of the indices in the train and test sets of each fold.
                Defaults to None.
            stratified_feature (str, optional): The feature name to preserve the percentage of samples for each class in
                the stratified folds. Defaults to None.
            hooks (Dict[str, Callable[[Model, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame], Any]], optional):
                Dictionary of callable hooks that are called within the CV loop. The callable retrieves the current trained
                modeld and the current CV folds in the following order: X_train, y_train, X_test, y_test. Defaults to {}.
            hook_kwargs (Dict[str, Dict[str, Any]], optional): Dictionary holding hook specefic keyword arguments.
                Defaults to {}.

        Returns:
            Tuple[CvResults, CvResults, Dict[str, List[Any]]]: First CvResults object reflects the training data,
                second CvResults object the test data, dictionary object holds the return values of the applied hooks.
        """
        if include_labcodes and "labcode" not in experiments.columns:
            raise ValueError("No labcodes available for the provided experiments.")

        if len(self.outputs) > 1:  # type: ignore
            raise NotImplementedError(
                "Cross validation not implemented for multi-output models"
            )

        if stratified_feature is not None:
            if stratified_feature not in (
                self.inputs.get_keys() + self.outputs.get_keys()  # type: ignore
            ):
                raise ValueError(
                    "The feature to be stratified is not in the model inputs or outputs"
                )
            try:
                feat = self.inputs.get_by_key(stratified_feature)  # type: ignore
            except KeyError:
                feat = self.outputs.get_by_key(stratified_feature)  # type: ignore
            if not isinstance(
                feat,
                (DiscreteInput, CategoricalInput, CategoricalOutput, ContinuousOutput),
            ):
                raise ValueError(
                    "The feature to be stratified needs to be a DiscreteInput, CategoricalInput, CategoricalOutput, or ContinuousOutput"
                )

        # first filter the experiments based on the model setting
        experiments = self._preprocess_experiments(experiments)
        n = len(experiments)
        if folds > n:
            warnings.warn(
                f"Training data only has {n} experiments, which is less than folds, fallback to LOOCV."
            )
            folds = n
        elif n == 0:
            raise ValueError("Experiments is empty.")
        elif folds < 2 and folds != -1:
            raise ValueError("Folds must be -1 for LOO, or > 1.")
        elif folds == -1:
            folds = n
        # preprocess hooks
        if hooks is None:
            hooks = {}
        if hook_kwargs is None:
            hook_kwargs = {}
        hook_results = {key: [] for key in hooks.keys()}

        # instantiate kfold object
        if stratified_feature is None:
            cv = KFold(n_splits=folds, shuffle=True, random_state=random_state)
            cv_func = cv.split(experiments)
        else:
            cv = StratifiedKFold(
                n_splits=folds, shuffle=True, random_state=random_state
            )
            cv_func = cv.split(
                experiments.drop([stratified_feature], axis=1),
                experiments[stratified_feature],
            )

        key = self.outputs.get_keys()[0]  # type: ignore
        train_results = []
        test_results = []
        # now get the indices for the split
        for train_index, test_index in cv_func:
            X_train = experiments.iloc[train_index][self.inputs.get_keys()]  # type: ignore
            X_test = experiments.iloc[test_index][self.inputs.get_keys()]  # type: ignore
            y_train = experiments.iloc[train_index][self.outputs.get_keys()]  # type: ignore
            y_test = experiments.iloc[test_index][self.outputs.get_keys()]  # type: ignore
            train_labcodes = (
                experiments.iloc[train_index]["labcode"] if include_labcodes else None
            )
            test_labcodes = (
                experiments.iloc[test_index]["labcode"] if include_labcodes else None
            )
            # now fit the model
            self._fit(X_train, y_train)
            # now do the scoring
            y_test_pred = self.predict(X_test)  # type: ignore
            y_train_pred = self.predict(X_train)  # type: ignore
            # now store the results
            train_results.append(
                CvResult(  # type: ignore
                    key=key,
                    observed=y_train[key],
                    predicted=y_train_pred[key + "_pred"],
                    standard_deviation=y_train_pred[key + "_sd"],
                    X=X_train if include_X else None,
                    labcodes=train_labcodes,
                )
            )
            test_results.append(
                CvResult(  # type: ignore
                    key=key,
                    observed=y_test[key],
                    predicted=y_test_pred[key + "_pred"],
                    standard_deviation=y_test_pred[key + "_sd"],
                    X=X_test if include_X else None,
                    labcodes=test_labcodes,
                )
            )
            # now call the hooks if available
            for hookname, hook in hooks.items():
                hook_results[hookname].append(
                    hook(
                        surrogate=self,  # type: ignore
                        X_train=X_train,
                        y_train=y_train,
                        X_test=X_test,
                        y_test=y_test,
                        **hook_kwargs.get(hookname, {}),
                    )
                )
        return (
            CvResults(results=train_results),
            CvResults(results=test_results),
            hook_results,
        )



import warnings
from typing import ClassVar, List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import validator

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.categorical import _CAT_SEP, CategoricalInput
from bofire.data_models.features.feature import Input
from bofire.data_models.molfeatures.api import (
    AnyMolFeatures,
    Fingerprints,
    FingerprintsFragments,
    Fragments,
    MordredDescriptors,
)
from bofire.utils.cheminformatics import smiles2mol


class MolecularInput(Input):
    type: Literal["MolecularInput"] = "MolecularInput"
    order: ClassVar[int] = 6

    @staticmethod
    def valid_transform_types() -> List[AnyMolFeatures]:
        return [Fingerprints, FingerprintsFragments, Fragments, MordredDescriptors]  # type: ignore

    def validate_experimental(
        self, values: pd.Series, strict: bool = False
    ) -> pd.Series:
        values = values.map(str)
        for smi in values:
            smiles2mol(smi)

        return values

    def validate_candidental(self, values: pd.Series) -> pd.Series:
        values = values.map(str)
        for smi in values:
            smiles2mol(smi)
        return values

    def is_fixed(self) -> bool:
        return False

    def fixed_value(self, transform_type: Optional[AnyMolFeatures] = None) -> None:
        return None

    def sample(self, n: int) -> pd.Series:
        raise ValueError("Sampling not supported for `MolecularInput`")

    def get_bounds(
        self, transform_type: AnyMolFeatures, values: pd.Series
    ) -> Tuple[List[float], List[float]]:
        if values is None:
            raise NotImplementedError(
                "`values` is currently required for `MolecularInput`"
            )
        else:
            data = self.to_descriptor_encoding(transform_type, values)

        lower = data.min(axis=0).values.tolist()
        upper = data.max(axis=0).values.tolist()

        return lower, upper

    def to_descriptor_encoding(
        self, transform_type: AnyMolFeatures, values: pd.Series
    ) -> pd.DataFrame:
        """Converts values to descriptor encoding.

        Args:
            values (pd.Series): Values to transform.

        Returns:
            pd.DataFrame: Descriptor encoded dataframe.
        """
        descriptor_values = transform_type.get_descriptor_values(values)

        descriptor_values.columns = [
            f"{self.key}{_CAT_SEP}{d}" for d in transform_type.get_descriptor_names()
        ]
        descriptor_values.index = values.index

        return descriptor_values


class CategoricalMolecularInput(CategoricalInput, MolecularInput):
    type: Literal["CategoricalMolecularInput"] = "CategoricalMolecularInput"
    order: ClassVar[int] = 7

    @validator("categories")
    def validate_smiles(cls, categories: Sequence[str]):
        """validates that categories are valid smiles. Note that this check can only
        be executed when rdkit is available.

        Args:
            categories (List[str]): List of smiles

        Raises:
            ValueError: when string is not a smiles

        Returns:
            List[str]: List of the smiles
        """
        # check on rdkit availability:
        try:
            smiles2mol(categories[0])
        except NameError:
            warnings.warn("rdkit not installed, categories cannot be validated.")
            return categories

        for cat in categories:
            smiles2mol(cat)
        return categories

    @staticmethod
    def valid_transform_types() -> List[Union[AnyMolFeatures, CategoricalEncodingEnum]]:
        return CategoricalInput.valid_transform_types() + [
            Fingerprints,
            FingerprintsFragments,
            Fragments,
            MordredDescriptors,  # type: ignore
        ]

    def get_bounds(
        self,
        transform_type: Union[CategoricalEncodingEnum, AnyMolFeatures],
        values: Optional[pd.Series] = None,
    ) -> Tuple[List[float], List[float]]:
        if isinstance(transform_type, CategoricalEncodingEnum):
            # we are just using the standard categorical transformations
            return super().get_bounds(transform_type=transform_type, values=values)
        else:
            # in case that values is None, we return the optimization bounds
            # else we return the complete bounds
            data = self.to_descriptor_encoding(
                transform_type=transform_type,
                values=pd.Series(self.get_allowed_categories())
                if values is None
                else pd.Series(self.categories),
            )
        lower = data.min(axis=0).values.tolist()
        upper = data.max(axis=0).values.tolist()
        return lower, upper

    def from_descriptor_encoding(
        self, transform_type: AnyMolFeatures, values: pd.DataFrame
    ) -> pd.Series:
        """Converts values back from descriptor encoding.

        Args:
            values (pd.DataFrame): Descriptor encoded dataframe.

        Raises:
            ValueError: If descriptor columns not found in the dataframe.

        Returns:
            pd.Series: Series with categorical values.
        """

        # This method is modified based on the categorical descriptor feature
        # TODO: move it to more central place
        cat_cols = [
            f"{self.key}{_CAT_SEP}{d}" for d in transform_type.get_descriptor_names()
        ]
        # we allow here explicitly that the dataframe can have more columns than needed to have it
        # easier in the backtransform.
        if np.any([c not in values.columns for c in cat_cols]):
            raise ValueError(
                f"{self.key}: Column names don't match categorical levels: {values.columns}, {cat_cols}."
            )
        s = pd.DataFrame(
            data=np.sqrt(
                np.sum(
                    (
                        values[cat_cols].to_numpy()[:, np.newaxis, :]
                        - self.to_descriptor_encoding(
                            transform_type=transform_type,
                            values=pd.Series(self.get_allowed_categories()),
                        ).to_numpy()
                    )
                    ** 2,
                    axis=2,
                )
            ),
            columns=self.get_allowed_categories(),
            index=values.index,
        ).idxmin(1)
        s.name = self.key
        return s



import numpy as np
import pandas as pd
import pytest

import bofire.data_models.outlier_detection.api as data_models
import bofire.outlier_detection.api as mapper
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.kernels.api import RBFKernel, ScaleKernel
from bofire.data_models.surrogates.api import ScalerEnum, SingleTaskGPSurrogate
from bofire.outlier_detection.outlier_detections import OutlierDetections


def neal_func(x):
    return 0.3 + 0.4 * x + 0.5 * np.sin(2.7 * x) + 1.1 / (1 + x**2)


np.random.seed(5)


def test_IterativeTrimming():
    # generate mock data
    n = 200
    noise = 0.2
    n_outlier = 60
    noise_outlier = 1

    x_ob = np.random.rand(n) * 6 - 3
    y_ob = neal_func(x_ob) + np.random.randn(n) * noise
    y_ob[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier
    )

    experiments = pd.DataFrame()
    experiments["x_1"] = x_ob
    experiments["y"] = y_ob
    experiments["valid_y"] = 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[0],  # type: ignore
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key=experiments.keys()[1])])  # type: ignore
    kernel = ScaleKernel(base_kernel=RBFKernel(ard=True))
    scaler = ScalerEnum.NORMALIZE
    ITGP_model = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs, outputs=outputs, kernel=kernel, scaler=scaler
        )
    )
    ITGP = mapper.map(ITGP_model)
    assert isinstance(ITGP.base_gp, SingleTaskGPSurrogate)  # type: ignore
    assert isinstance(ITGP, mapper.IterativeTrimming)
    # detect
    experiments1 = ITGP.detect(experiments=experiments)
    assert len(experiments[experiments["valid_y"] == 1]) != len(
        experiments1[experiments1["valid_y"] == 1]
    )
    assert len(experiments1[experiments1["valid_y"] == 0]) <= n / 2


def test_OutlierDetections():
    # generate mock data
    # test for single detector
    n = 200
    noise = 0.2
    n_outlier = 60
    noise_outlier = 1
    noise_outlier1 = 1.5
    x_ob = np.random.rand(n) * 6 - 3
    y_ob = neal_func(x_ob) + np.random.randn(n) * noise
    y_ob[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier
    )
    experiments = pd.DataFrame()
    experiments["x_1"] = x_ob
    experiments["y"] = y_ob
    experiments["valid_y"] = 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[0],  # type: ignore
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key=experiments.keys()[1])])  # type: ignore
    kernel = ScaleKernel(base_kernel=RBFKernel(ard=True))
    scaler = ScalerEnum.NORMALIZE
    ITGP_model1 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs, outputs=outputs, kernel=kernel, scaler=scaler
        )
    )
    ITGP = data_models.OutlierDetections(detectors=[ITGP_model1])
    ITGP = OutlierDetections(data_model=ITGP)
    assert isinstance(ITGP.detectors[0], mapper.IterativeTrimming)
    # detect
    experiments1 = ITGP.detect(experiments=experiments)
    assert len(experiments[experiments["valid_y"] == 1]) != len(
        experiments1[experiments1["valid_y"] == 1]
    )
    assert len(experiments1[experiments1["valid_y"] == 0]) <= n / 2
    # multiple detectors
    n = 200
    noise = 0.2
    n_outlier = 60
    noise_outlier = 1
    noise_outlier1 = 1.5
    x_ob = np.random.rand(n) * 6 - 3
    y_ob = neal_func(x_ob) + np.random.randn(n) * noise
    y_ob[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier
    )
    y_ob1 = y_ob
    y_ob1[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier1
    )
    experiments = pd.DataFrame()
    experiments["x_1"] = x_ob
    experiments["y"] = y_ob
    experiments["y1"] = y_ob1
    experiments["valid_y"] = 1
    experiments["valid_y1"] = 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[0],  # type: ignore
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    outputs1 = Outputs(features=[ContinuousOutput(key=experiments.keys()[1])])  # type: ignore
    outputs2 = Outputs(features=[ContinuousOutput(key=experiments.keys()[2])])  # type: ignore
    kernel = ScaleKernel(base_kernel=RBFKernel(ard=True))
    scaler = ScalerEnum.NORMALIZE
    ITGP_model1 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs, outputs=outputs1, kernel=kernel, scaler=scaler
        )
    )
    ITGP_model2 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs, outputs=outputs2, kernel=kernel, scaler=scaler
        )
    )
    ITGP = data_models.OutlierDetections(detectors=[ITGP_model1, ITGP_model2])
    ITGP = OutlierDetections(data_model=ITGP)
    assert isinstance(ITGP.detectors[0], mapper.IterativeTrimming)
    assert isinstance(ITGP.detectors[1], mapper.IterativeTrimming)
    # detect
    experiments1 = ITGP.detect(experiments=experiments)
    assert len(experiments[experiments["valid_y"] == 1]) != len(
        experiments1[experiments1["valid_y"] == 1]
    )
    assert len(experiments[experiments["valid_y1"] == 1]) != len(
        experiments1[experiments1["valid_y1"] == 1]
    )
    assert len(experiments1[experiments1["valid_y"] == 0]) <= n / 2
    assert len(experiments1[experiments1["valid_y1"] == 0]) <= n / 2
    # different inputs
    n = 200
    noise = 0.2
    n_outlier = 60
    noise_outlier = 1
    noise_outlier1 = 1.5
    x_ob = np.random.rand(n) * 6 - 3
    x_ob1 = np.random.randn(n) * 6 - 3
    y_ob = neal_func(x_ob) + np.random.randn(n) * noise
    y_ob[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier
    )
    y_ob1 = y_ob
    y_ob1[:n_outlier] = (
        neal_func(x_ob1[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier1
    )
    experiments = pd.DataFrame()
    experiments["x"] = x_ob
    experiments["x1"] = x_ob
    experiments["y"] = y_ob
    experiments["y1"] = y_ob1
    experiments["valid_y"] = 1
    experiments["valid_y1"] = 1
    inputs1 = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[0],  # type: ignore
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    inputs2 = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[1],  # type: ignore
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    outputs1 = Outputs(features=[ContinuousOutput(key=experiments.keys()[2])])  # type: ignore
    outputs2 = Outputs(features=[ContinuousOutput(key=experiments.keys()[3])])  # type: ignore

    kernel = ScaleKernel(base_kernel=RBFKernel(ard=True))
    scaler = ScalerEnum.NORMALIZE
    ITGP_model1 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs1, outputs=outputs1, kernel=kernel, scaler=scaler
        )
    )
    ITGP_model2 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs2, outputs=outputs2, kernel=kernel, scaler=scaler
        )
    )
    ITGP = data_models.OutlierDetections(detectors=[ITGP_model1, ITGP_model2])
    ITGP = OutlierDetections(data_model=ITGP)
    assert isinstance(ITGP.detectors[0], mapper.IterativeTrimming)
    assert isinstance(ITGP.detectors[1], mapper.IterativeTrimming)
    # detect
    experiments1 = ITGP.detect(experiments=experiments)
    assert len(experiments[experiments["valid_y"] == 1]) != len(
        experiments1[experiments1["valid_y"] == 1]
    )
    assert len(experiments[experiments["valid_y1"] == 1]) != len(
        experiments1[experiments1["valid_y1"] == 1]
    )
    assert len(experiments1[experiments1["valid_y"] == 0]) <= n / 2
    assert len(experiments1[experiments1["valid_y1"] == 0]) <= n / 2


def test_outlier_detectors_check_compatibility():
    data_model1 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=Inputs(
                features=[
                    ContinuousInput(
                        key=f"x_{i+1}",
                        bounds=(-4, 4),
                    )
                    for i in range(3)
                ]
                + [
                    CategoricalDescriptorInput(
                        key="cat",
                        categories=["apple", "banana"],
                        descriptors=["length", "width"],
                        values=[[1, 2], [3, 4]],
                    )
                ]
            ),
            outputs=Outputs(features=[ContinuousOutput(key="y")]),
            scaler=ScalerEnum.NORMALIZE,
            input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
        )
    )
    data_model2 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=Inputs(
                features=[
                    ContinuousInput(
                        key=f"x_{i+1}",
                        bounds=(-4, 4),
                    )
                    for i in range(2)
                ]
                + [
                    CategoricalDescriptorInput(
                        key="cat",
                        categories=["apple", "banana"],
                        descriptors=["length", "width"],
                        values=[[1, 2], [3, 4]],
                    )
                ]
            ),
            outputs=Outputs(features=[ContinuousOutput(key="y2")]),
            scaler=ScalerEnum.NORMALIZE,
            input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
        )
    )
    data_model = data_models.OutlierDetections(detectors=[data_model1, data_model2])
    # models = OutlierDetections(data_model=data_model)
    # check too less input features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(
        ValueError, match=r"Model with index \d+ has more features than acceptable."
    ):
        data_model._check_compability(inp, out)
    # check unused input features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(4)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(ValueError):
        data_model._check_compability(inp, out)
    # check wrong input feature
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(4)
        ]
        + [
            ContinuousInput(
                key="cat",
                bounds=(-4, 4),
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(ValueError, match=r"Features with key \w+ are incompatible."):
        data_model._check_compability(inp, out)
    # check too less output features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y")])
    with pytest.raises(ValueError, match="Output features do not match."):
        data_model._check_compability(inp, out)
    # check too many output features
    out = Outputs(
        features=[
            ContinuousOutput(key="y"),
            ContinuousOutput(key="y2"),
            ContinuousOutput(key="y3"),
        ]
    )
    with pytest.raises(ValueError, match="Output features do not match."):
        data_model._check_compability(inp, out)
    # check wrong output features
    out = Outputs(
        features=[
            ContinuousOutput(key="y"),
            ContinuousOutput(key="y3"),
        ]
    )
    with pytest.raises(ValueError, match="Output features do not match."):
        data_model._check_compability(inp, out)
    # check valid
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    data_model._check_compability(inp, out)


def test_outlier_detectors_unique_outputs():
    data_model1 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(3)
                ]
            ),
            outputs=Outputs(features=[ContinuousOutput(key="y")]),
            scaler=ScalerEnum.NORMALIZE,
        )
    )
    data_model2 = data_models.IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(2)
                ]
                + [CategoricalInput(key="x_3", categories=["apple", "banana"])]
            ),
            outputs=Outputs(features=[ContinuousOutput(key="y")]),
            scaler=ScalerEnum.NORMALIZE,
        )
    )
    with pytest.raises(
        ValueError, match="Output feature keys are not unique across detectors."
    ):
        data_models.OutlierDetections(detectors=[data_model1, data_model2])



from abc import abstractmethod
from typing import Callable, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import Field, PositiveFloat
from scipy.stats import norm, uniform
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Domain


class OutlierPrior(BaseModel):
    type: str


class UniformOutlierPrior(OutlierPrior):
    type: Literal["UniformOutlierPrior"] = "UniformOutlierPrior"
    bounds: Tuple[float, float]

    def sample(self, n_samples: int) -> np.ndarray:
        return uniform(
            self.bounds[0],
            self.bounds[1] - self.bounds[0],
        ).rvs(n_samples)


class NormalOutlierPrior(OutlierPrior):
    type: Literal["NormalOutlierPrior"] = "NormalOutlierPrior"
    loc: float
    scale: PositiveFloat

    def sample(self, n_samples: int) -> np.ndarray:
        return norm(self.loc, self.scale).rvs(n_samples)


AnyOutlierPrior = Union[UniformOutlierPrior, NormalOutlierPrior]


class Benchmark:
    def __init__(
        self,
        outlier_rate: Annotated[float, Field(ge=0, lt=1)] = 0,
        outlier_prior: Optional[AnyOutlierPrior] = None,
    ):
        self.outlier_rate = outlier_rate
        self.outlier_prior = outlier_prior

    def f(
        self,
        candidates: pd.DataFrame,
        return_complete: bool = False,
    ) -> pd.DataFrame:
        Y = self._f(candidates)
        if self.outlier_prior is not None:
            for output_feature in self.domain.outputs.get_keys():
                # no_outliers = int(len(Y) * self.outlier_rate)
                ix2 = np.zeros(len(Y), dtype=bool)
                ix1 = uniform().rvs(len(Y))
                # ix2[np.random.choice(len(Y), no_outliers, replace=False)] = True
                ix2 = ix1 <= self.outlier_rate
                n_outliers = sum(ix2)
                Y.loc[ix2, output_feature] = Y.loc[ix2, output_feature] + self.outlier_prior.sample(n_outliers)  # type: ignore
        if return_complete:
            return pd.concat([candidates, Y], axis=1)

        return Y

    @abstractmethod
    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        pass

    def get_optima(self) -> pd.DataFrame:
        raise NotImplementedError()

    @property
    def domain(self) -> Domain:
        return self._domain  # type: ignore


class GenericBenchmark(Benchmark):
    def __init__(
        self,
        domain: Domain,
        func: Callable[[pd.DataFrame], pd.DataFrame],
        outlier_rate: Annotated[float, Field(ge=0, lt=1)] = 0,
        outlier_prior: Optional[AnyOutlierPrior] = None,
    ):
        super().__init__(outlier_prior=outlier_prior, outlier_rate=outlier_rate)
        self._domain = domain
        self.func = func

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        return self.func(candidates)



from pydantic import confloat

from bofire.data_models.base import BaseModel


class PredictedValue(BaseModel):
    """Container holding information regarding individual predictions.

    Used to comunicate with backend services.

    Attributes:
        predictedValue (float): The predicted value.
        standardDeviation (float): The standard deviation associated with the prediction.
            Has to be greater/equal than zero.
    """

    predictedValue: float
    standardDeviation: confloat(ge=0)  # type: ignore



from pydantic import validator

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    MolecularInput,
    NumericalInput,
)
from bofire.data_models.molfeatures.api import Fingerprints, MolFeatures
from bofire.data_models.surrogates.surrogate import Surrogate


class BotorchSurrogate(Surrogate):
    @validator("input_preprocessing_specs", always=True)
    def validate_input_preprocessing_specs(cls, v, values):
        inputs = values["inputs"]
        categorical_keys = inputs.get_keys(CategoricalInput, exact=True)
        descriptor_keys = inputs.get_keys(CategoricalDescriptorInput, exact=True)
        molecular_keys = inputs.get_keys(MolecularInput, exact=True)
        for key in categorical_keys:
            if (
                v.get(key, CategoricalEncodingEnum.ONE_HOT)
                != CategoricalEncodingEnum.ONE_HOT
            ):
                raise ValueError(
                    "Botorch based models have to use one hot encodings for categoricals"
                )
            else:
                v[key] = CategoricalEncodingEnum.ONE_HOT
        # TODO: include descriptors into probabilistic reparam via OneHotToDescriptor input transform
        for key in descriptor_keys:
            if v.get(key, CategoricalEncodingEnum.DESCRIPTOR) not in [
                CategoricalEncodingEnum.DESCRIPTOR,
                CategoricalEncodingEnum.ONE_HOT,
            ]:
                raise ValueError(
                    "Botorch based models have to use one hot encodings or descriptor encodings for categoricals."
                )
            elif v.get(key) is None:
                v[key] = CategoricalEncodingEnum.DESCRIPTOR
        for key in inputs.get_keys(NumericalInput):
            if v.get(key) is not None:
                raise ValueError(
                    "Botorch based models have to use internal transforms to preprocess numerical features."
                )
        # TODO: include descriptors into probabilistic reparam via OneHotToDescriptor input transform
        for key in molecular_keys:
            mol_encoding = v.get(key, Fingerprints())
            if not isinstance(mol_encoding, MolFeatures):
                raise ValueError(
                    "Botorch based models have to use fingerprints, fragments, fingerprints_fragments, or molecular descriptors for molecular inputs"
                )
            v[key] = mol_encoding
        return v



import numpy as np
import pandas as pd
import pytest
from pydantic.error_wrappers import ValidationError

import tests.bofire.data_models.specs.api as specs
from bofire.data_models.constraints.api import (
    Constraint,
    LinearConstraint,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.domain.api import Constraints, Inputs
from bofire.data_models.enum import SamplingMethodEnum
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput


def test_from_greater_equal():
    spec = specs.constraints.valid(LinearInequalityConstraint).spec()
    c = LinearInequalityConstraint.from_greater_equal(
        features=spec["features"],
        coefficients=spec["coefficients"],
        rhs=spec["rhs"],
    )
    assert c.rhs == spec["rhs"] * -1.0
    assert c.coefficients == [-1.0 * coef for coef in spec["coefficients"]]
    assert c.features == spec["features"]


def test_as_greater_equal():
    spec = specs.constraints.valid(LinearInequalityConstraint).spec()
    c = LinearInequalityConstraint.from_greater_equal(
        features=spec["features"],
        coefficients=spec["coefficients"],
        rhs=spec["rhs"],
    )
    features, coefficients, rhs = c.as_greater_equal()
    assert c.rhs == rhs * -1.0
    assert coefficients == [-1.0 * coef for coef in c.coefficients]
    assert c.features == features


def test_from_smaller_equal():
    spec = specs.constraints.valid(LinearInequalityConstraint).spec()
    c = LinearInequalityConstraint.from_smaller_equal(
        features=spec["features"],
        coefficients=spec["coefficients"],
        rhs=spec["rhs"],
    )
    assert c.rhs == spec["rhs"]
    assert c.coefficients == spec["coefficients"]
    assert c.features == spec["features"]


def test_as_smaller_equal():
    spec = specs.constraints.valid(LinearInequalityConstraint).spec()
    c = LinearInequalityConstraint.from_smaller_equal(
        features=spec["features"],
        coefficients=spec["coefficients"],
        rhs=spec["rhs"],
    )
    features, coefficients, rhs = c.as_smaller_equal()
    assert c.rhs == rhs
    assert coefficients == c.coefficients
    assert c.features == features


# test the Constraints Class
c1 = specs.constraints.valid(LinearEqualityConstraint).obj()
c2 = specs.constraints.valid(LinearInequalityConstraint).obj()
c3 = specs.constraints.valid(NChooseKConstraint).obj()
c4 = specs.constraints.valid(NonlinearEqualityConstraint).obj()
c5 = specs.constraints.valid(NonlinearInequalityConstraint).obj()
c6 = LinearInequalityConstraint.from_smaller_equal(
    features=["f1", "f2", "f3"], coefficients=[1, 1, 1], rhs=100.0
)

if1 = ContinuousInput(key="f1", bounds=(0, 2))
if2 = ContinuousInput(key="f2", bounds=(0, 4))
if3 = ContinuousInput(key="f3", bounds=(3, 8))

inputs = Inputs(features=[if1, if2, if3])

constraints = Constraints(constraints=[c1, c2])
constraints2 = Constraints(constraints=[c4, c5])
constraints3 = Constraints(constraints=[c6])
constraints4 = Constraints(constraints=[c3])


@pytest.mark.parametrize(
    "constraints",
    [
        (["s"]),
        ([specs.constraints.valid(LinearInequalityConstraint).obj()], 5),
        (
            [specs.constraints.valid(LinearInequalityConstraint).obj()],
            ContinuousOutput(key="s"),
        ),
    ],
)
def test_constraints_invalid_constraint(constraints):
    with pytest.raises((ValueError, TypeError, KeyError, ValidationError)):
        Constraints(constraints=constraints)


@pytest.mark.parametrize(
    "constraints, ConstraintType, exact, expected",
    [
        (constraints + constraints4, LinearConstraint, True, []),
        (constraints + constraints4, LinearConstraint, False, [c1, c2]),
        (constraints + constraints4, Constraint, False, [c1, c2, c3]),
        (constraints + constraints4, NChooseKConstraint, False, [c3]),
    ],
)
def test_constraints_get(constraints, ConstraintType, exact, expected):
    returned = constraints.get(ConstraintType, exact=exact).constraints
    assert returned == expected
    for i in range(len(expected)):
        assert id(expected[i]) == id(returned[i])


def test_constraints_plus():
    returned = constraints + constraints4 + constraints2
    assert returned.constraints == [c1, c2, c3, c4, c5]


@pytest.mark.parametrize(
    "constraints, num_candidates",
    [
        (constraints2, 5),
        (constraints4, 5),
    ],
)
def test_constraints_call(constraints, num_candidates):
    candidates = inputs.sample(num_candidates, SamplingMethodEnum.UNIFORM)
    returned = constraints(candidates)
    assert returned.shape == (num_candidates, len(constraints))


@pytest.mark.parametrize(
    "constraints, num_candidates, fulfilled",
    [
        (constraints2, 5, False),
        (constraints3, 5, True),
    ],
)
def test_constraints_is_fulfilled(constraints, num_candidates, fulfilled):
    candidates = inputs.sample(num_candidates, SamplingMethodEnum.UNIFORM)
    returned = constraints.is_fulfilled(candidates)
    assert returned.shape == (num_candidates,)
    assert returned.dtype == bool
    assert returned.all() == fulfilled


@pytest.mark.parametrize(
    "constraints, num_candidates",
    [
        (constraints, 2),
        (constraints2, 2),
    ],
)
def test_constraints_jacobian(constraints, num_candidates):
    candidates = inputs.sample(num_candidates, SamplingMethodEnum.UNIFORM)
    returned = constraints.jacobian(candidates)
    assert np.all(
        [
            list(returned[i].columns) == ["dg/df1", "dg/df2", "dg/df3"]
            for i, c in enumerate(constraints)
        ]
    )
    assert np.all(
        [
            returned[i].shape == (num_candidates, len(inputs))
            for i, c in enumerate(constraints)
        ]
    )
    for i, c in enumerate(constraints):
        if isinstance(c, LinearConstraint):
            assert np.allclose(
                returned[i],
                np.tile(
                    c.coefficients / np.linalg.norm(c.coefficients), (num_candidates, 1)
                ),
            )
        if isinstance(c, NonlinearConstraint):
            res = candidates.eval(c.jacobian_expression)
            for j, col in enumerate(res):
                if not hasattr(col, "__iter__"):
                    res[j] = pd.Series(np.repeat(col, candidates.shape[0]))
            assert np.allclose(returned[i], pd.DataFrame(res).transpose())



import importlib
import random

import bofire.data_models.molfeatures.api as molfeatures
import bofire.data_models.molfeatures.names as names
from tests.bofire.data_models.specs.specs import Specs

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None
MORDRED_AVAILABLE = importlib.util.find_spec("mordred") is not None

specs = Specs([])

specs.add_valid(
    molfeatures.Fingerprints,
    lambda: {
        "bond_radius": random.randrange(1, 6),
        "n_bits": random.randrange(32, 2048),
    },
)

if RDKIT_AVAILABLE:
    specs.add_valid(
        molfeatures.Fragments,
        lambda: {
            "fragments": random.sample(
                names.fragments, k=random.randrange(1, len(names.fragments))
            )
        },
    )
    specs.add_valid(
        molfeatures.FingerprintsFragments,
        lambda: {
            "bond_radius": random.randrange(1, 6),
            "n_bits": random.randrange(32, 2048),
            "fragments": random.sample(
                names.fragments, k=random.randrange(1, len(names.fragments))
            ),
        },
    )

    if MORDRED_AVAILABLE:
        specs.add_valid(
            molfeatures.MordredDescriptors,
            lambda: {
                "descriptors": random.sample(names.mordred, k=random.randrange(1, 10))
            },
        )



import collections.abc as collections
from typing import Any, Callable, List, Sequence, Type, Union, get_args, get_origin

import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra


class BaseModel(PydanticBaseModel):
    class Config:
        validate_assignment = True
        arbitrary_types_allowed = False
        copy_on_model_validation = "none"
        extra = Extra.forbid

        json_encoders = {
            pd.DataFrame: lambda x: x.to_dict(orient="list"),
            pd.Series: lambda x: x.to_list(),
        }


def filter_by_attribute(
    data: Sequence,
    attribute_getter: Callable[[Type], Any],
    includes: Union[Type, Sequence[Type]] = None,
    excludes: Union[Type, Sequence[Type]] = None,
    exact: bool = False,
) -> List:
    """Returns those data elements where the attribute is of one of the include types.

    Args:
        data: to be filtered
        attribute_getter: expects an item of the data list and returns the attribute to filter by
        includes: attribute types that should be kept, sub-type are included by default, see exact
        excludes: attribute types that will be excluded even if they are sub-types of or include types.
        exact: true for not including subtypes

    Returns:
        list of data point with attributes as filtered for
    """
    data_with_attr = []
    for d in data:
        try:
            attribute_getter(d)
            data_with_attr.append(d)
        except AttributeError:
            pass

    filtered = filter_by_class(
        data_with_attr,
        includes=includes,
        excludes=excludes,
        exact=exact,
        key=attribute_getter,
    )
    return filtered


def filter_by_class(
    data: Sequence,
    includes: Union[Type, Sequence[Type]] = None,
    excludes: Union[Type, Sequence[Type]] = None,
    exact: bool = False,
    key: Callable[[Type], Any] = lambda x: x,
) -> List:
    """Returns those data elements where are one of the include types.

    Args:
        data: to be filtered
        includes: attribute types that should be kept, sub-type are included by default, see exact
        excludes: attribute types that will be excluded even if they are sub-types of or include types.
        exact: true for not including subtypes
        key: maps a data list item to something that is used for filtering, identity by default

    Returns:
        filtered list of data points
    """
    if includes is None:
        includes = []
    if not isinstance(includes, collections.Sequence):
        includes = [includes]
    if excludes is None:
        excludes = []
    if not isinstance(excludes, collections.Sequence):
        excludes = [excludes]

    if len(includes) == len(excludes) == 0:
        raise ValueError("no filter provided")

    if len(includes) == 0:
        includes = [object]

    includes_ = []
    for incl in includes:
        if get_origin(incl) is Union:
            includes_ += get_args(incl)
        else:
            includes_.append(incl)
    includes = includes_
    excludes_ = []
    for excl in excludes:
        if get_origin(excl) is Union:
            excludes_ += get_args(excl)
        else:
            excludes_.append(excl)
    excludes = excludes_

    if len([x for x in includes if x in excludes]) > 0:
        raise ValueError("includes and excludes overlap")

    if exact:
        return [
            d for d in data if type(key(d)) in includes and type(key(d)) not in excludes
        ]
    return [
        d
        for d in data
        if isinstance(key(d), tuple(includes))  # type: ignore
        and not isinstance(key(d), tuple(excludes))  # type: ignore
    ]



import numpy as np
import pandas as pd
import pytest

from bofire.data_models.features.api import ContinuousOutput
from bofire.data_models.objectives.api import MaximizeSigmoidObjective
from bofire.plot.objective import plot_objective_plotly


@pytest.mark.parametrize(
    "feature, data",
    [
        (
            ContinuousOutput(
                key="of1", objective=MaximizeSigmoidObjective(w=1, tp=15, steepness=0.5)
            ),
            None,
        ),
        (
            ContinuousOutput(
                key="of1", objective=MaximizeSigmoidObjective(w=1, tp=15, steepness=0.5)
            ),
            pd.DataFrame(
                columns=["of1", "of2", "of3"],
                index=range(5),
                data=np.random.uniform(size=(5, 3)),
            ),
        ),
    ],
)
def test_output_feature_plot(feature, data):
    print(feature)
    plot_objective_plotly(
        feature=feature,
        lower=0,
        upper=30,
        values=data["of1"] if data is not None else None,
    )



import pandas as pd

import bofire.strategies.api as strategies
import bofire.strategies.mapper as strategy_mapper
from bofire.benchmarks.multi import ZDT1
from bofire.data_models.domain.api import Domain
from bofire.data_models.strategies.api import (
    QparegoStrategy as QparegoStrategyDataModel,
)
from bofire.data_models.strategies.api import (
    RejectionSampler as RejectionSamplerDataModel,
)
from bofire.runners.api import run
from bofire.utils.multiobjective import compute_hypervolume


def test_benchmark():
    zdt1 = ZDT1(n_inputs=5)
    qparego_factory = QparegoStrategyDataModel

    n_initial_samples = 10
    n_runs = 3
    n_iterations = 2

    def sample(domain):
        nonlocal n_initial_samples
        sampler = strategies.map(RejectionSamplerDataModel(domain=domain))
        sampled = sampler.ask(n_initial_samples)

        return sampled

    def hypervolume(domain: Domain, experiments: pd.DataFrame) -> float:
        return compute_hypervolume(domain, experiments, ref_point={"y1": 10, "y2": 10})

    results = run(
        zdt1,
        strategy_factory=lambda domain: strategy_mapper.map(
            qparego_factory(domain=domain)
        ),
        n_iterations=n_iterations,
        metric=hypervolume,
        initial_sampler=sample,
        n_runs=n_runs,
        n_procs=1,
    )

    assert len(results) == n_runs
    for experiments, best in results:
        assert experiments is not None
        assert experiments.shape[0] == n_initial_samples + n_iterations
        assert best.shape[0] == n_iterations
        assert isinstance(best, pd.Series)
        assert isinstance(experiments, pd.DataFrame)



from bofire.data_models.priors.api import BOTORCH_LENGTHCALE_PRIOR
from bofire.plot.api import plot_prior_pdf_plotly


def test_plot_prior_pdf_plotly():
    plot_prior_pdf_plotly(BOTORCH_LENGTHCALE_PRIOR(), lower=0, upper=10)



import numpy as np
import pandas as pd
import pytest

from bofire.data_models.domain.domain import is_numeric


@pytest.mark.parametrize(
    "df, expected",
    [
        (
            pd.DataFrame(
                {
                    "col": [1, 2, 10, np.nan, "a"],
                    "col2": ["a", 10, 30, 40, 50],
                    "col3": [1, 2, 3, 4, 5.0],
                }
            ),
            False,
        ),
        (
            pd.DataFrame(
                {
                    "col": [1, 2, 10, np.nan, 6],
                    "col2": [5, 10, 30, 40, 50],
                    "col3": [1, 2, 3, 4, 5.0],
                }
            ),
            False,
        ),
        (
            pd.DataFrame(
                {
                    "col": [1, 2, 10, 7.0, 6],
                    "col2": [5, 10, 30, 40, 50],
                    "col3": [1, 2, 3, 4, 5.0],
                }
            ),
            True,
        ),
        (
            pd.Series([1, 2, 10, 7.0, 6]),
            True,
        ),
        (
            pd.Series([1, 2, "abc", 7.0, 6]),
            False,
        ),
    ],
)
def test_is_numeric(df, expected):
    assert is_numeric(df) == expected



from enum import Enum


class OptimalityCriterionEnum(Enum):
    D_OPTIMALITY = "D_OPTIMALITY"
    E_OPTIMALITY = "E_OPTIMALITY"
    A_OPTIMALITY = "A_OPTIMALITY"
    G_OPTIMALITY = "G_OPTIMALITY"
    K_OPTIMALITY = "K_OPTIMALITY"
    SPACE_FILLING = "SPACE_FILLING"



import logging
import os
from typing import Callable, Dict, List, Optional

import pandas as pd

from bofire.benchmarks.benchmark import Benchmark
from bofire.data_models.domain.api import Domain

# Create a folder for the log file, if not alredy exists.
if not os.path.exists("bofire_logs"):
    os.makedirs("bofire_logs")

# Create a Logger that safes console output.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
file_handler = logging.FileHandler("bofire_logs/aspen_benchmark.log")
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)


class Aspen_benchmark(Benchmark):
    """This class connects to a Aspen plus file that runs the desired process.
    It writes incoming input values into Aspen plus, runs the simulation and returns the results.
    When initializing this class, make sure not to block multiple Aspen plus licenses at once
    when is not absolutely needed.

    Args:
        Benchmark: Subclass of the Benchmark function class.
    """

    def __init__(
        self,
        filename: str,
        domain: Domain,
        paths: Dict[str, str],
        additional_output_keys: Optional[List] = None,
        translate_into_aspen_readable: Optional[
            Callable[[Domain, pd.DataFrame], pd.DataFrame]
        ] = None,
        **kwargs,
    ) -> None:
        """Initializes Aspen_benchmark. A class that connects to Aspen plus.

        Args:
            filename (str): Filepath of the Aspen plus simulation file.
            domain (Domain): Domain of the benchmark setting inclunding bounds and information about input values.
            paths (dict[str, str]): A dictionary with the key value pairs "key_of_variable": "path_to_variable".
            The keys must be the same as provided in the domain.
            translate_into_aspen_readable (Optional: Callable): A function that converts the columns of a candidate dataframe into
            integers or floats so Aspen plus is able to read their values.

        Raises:
            ValueError: In case the number of provided variable names does not match the number of provided Aspen variable tree paths.
        """
        super().__init__(**kwargs)
        if os.path.exists(filename):
            self.filename = filename
        else:
            raise ValueError("Unable to find Aspen file " + filename)

        self.translate_into_aspen_readable = translate_into_aspen_readable
        self._domain = domain
        self.additional_output_keys = additional_output_keys or []

        for key in self.domain.get_feature_keys():
            # Check, if every input and output variable has a path to Aspen provided.
            if key not in paths.keys():
                raise ValueError("Path for " + key + " is not provided.")

        self.paths = paths
        self.aspen_is_running = False

    # Start Aspen
    def start_aspen(self):
        """Starts Aspen plus and opens desired simulation file.

        Raises:
            ValueError: In case it is not possible to start Aspen plus.
        """
        import win32com.client as win32  # type: ignore

        logger.info("Starting Aspen plus")
        # Aspen should be accessible from every function as a global variable.
        global aspen
        try:
            aspen = win32.Dispatch("Apwn.Document")
            aspen.InitFromFile2(os.path.abspath(self.filename))
            self.aspen_is_running = True
        except OSError as e:
            logger.exception(e)
            raise ValueError(e)

    # Run simulation in Aspen
    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        """Function evaluation of the Aspen plus benchmark. Passes input values to Aspen, runs
        a simulation and reads the output values from Aspen. Checks for errors/warnings from Aspen.

        Args:
            candidates (pd.DataFrame): Input values that need to be passed to Aspen.
            Function can take either one input vector as a row in the input dataframe or multiple input vectors.

        Returns:
            pd.DataFrame: Output values from Aspen. The dataframe includes valid_(variable_name) columns for
            each output variable when the simulation went successful.
        """

        # Only start Aspen, when it is not already blocking.
        if self.aspen_is_running is False:
            self.start_aspen()

        # Make inputs Aspen-readable
        if self.translate_into_aspen_readable is not None:
            X = self.translate_into_aspen_readable(
                domain=self.domain,  # type: ignore
                candidates=candidates.copy(),
            )
        else:
            X = candidates

        y_outputs = {
            k: []
            for key in self.domain.outputs.get_keys()
            for k in (key, "valid_" + key)
        }
        add_outputs = {key: [] for key in self.additional_output_keys}

        # Iterate through dataframe rows to retrieve multiple input vectors. Running seperate simulations for each.
        for index, row in X.iterrows():
            logger.info("Writing inputs into Aspen")
            # Write input variables corresping to columns into aspen according to predefined paths.
            for key in self.domain.inputs.get_keys():
                try:
                    aspen.Tree.FindNode(self.paths.get(key)).Value = row[key]
                except ConnectionAbortedError:
                    logger.exception("Not able to write " + key + " into Aspen.")
                    raise ValueError("Not able to write " + key + " into Aspen.")

            # Reset Aspen simulation
            aspen.Reinit()
            # Start new Aspen simulation
            logger.info("Aspen simulation run " + str(index))
            aspen.Engine.Run2()
            logger.info("Simulation done.")

            # Retrieve outputs from Aspen and write into data frame
            logger.info("Retrieving outputs from Aspen.")
            try:
                # Check for errors during simulation in Aspen that disqualify the results
                status = aspen.Tree.FindNode(
                    "\\Data\\Results Summary\\Run-Status\\Output\\UOSSTAT2"
                ).Value

                if status != 8:
                    if status == 9:
                        logger.error(
                            "Result"
                            + " does not converge. Simulation status: "
                            + str(status)
                        )
                    elif status == 10:
                        logger.warning(
                            "Result"
                            + " gives an Aspen warning. Simulation status: "
                            + str(status)
                        )
                    else:
                        logger.warning("Unknown simulation status: " + str(status))

                for key in self.domain.outputs.get_keys():
                    y_outputs[key].append(
                        aspen.Tree.FindNode(self.paths.get(key)).Value
                    )
                    if status == 8:
                        # Result is valid and add valid_var = 1
                        # Status = 8 corresponds to a valid result that should be kept, 10 is a warning, 9 does not converge
                        y_outputs[f"valid_{key}"].append(1)
                    else:
                        y_outputs[f"valid_{key}"].append(0)

                for key in self.additional_output_keys:
                    add_outputs[key].append(
                        aspen.Tree.FindNode(self.paths.get(key)).Value
                    )

            except ConnectionAbortedError:
                logger.exception("Not able to retrieve values from Aspen.")
                raise ValueError("Not able to retrieve values from Aspen.")

        Y = pd.DataFrame(y_outputs)
        Z = pd.DataFrame(add_outputs)
        XYZ = pd.concat([candidates, Y, Z], axis=1)
        YZ = pd.concat([Y, Z], axis=1)
        logger.info("Simluation completed. Results:")
        logger.info(XYZ)
        return YZ



from typing import Dict, Optional

import pandas as pd

from bofire.data_models.base import BaseModel
from bofire.strategies.data_models.values import InputValue, OutputValue


class Candidate(BaseModel):
    """A Bofire candidate.

    Attributes:
        inputValues (Dict[str, InputValue]): Dictionary of input values where the key is the
            corresponding input feature key.
        outputValues (Dict[str, OutputValue], optional): Dictionary of output values where
            the key is the corresponding output feature key.
    """

    inputValues: Dict[str, InputValue]
    outputValues: Optional[Dict[str, OutputValue]] = None

    def to_series(self) -> pd.Series:
        """Transform to pandas series.

        Returns:
            pd.Series: pandas series which corresponds to one row in the original candidates dataframe
        """
        data = []
        index = []
        for key, value in self.inputValues.items():
            data.append(value.value)
            index.append(key)
        if self.outputValues is not None:
            for key, value in self.outputValues.items():
                data += [value.predictedValue, value.standardDeviation, value.objective]
                index += [f"{key}_{p}" for p in ["pred", "sd", "des"]]
        return pd.Series(data=data, index=index)



import bofire.data_models.outlier_detection.api as models
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate
from tests.bofire.data_models.specs.specs import Specs
from tests.bofire.data_models.specs.surrogates import specs as surrogates

specs = Specs([])

specs.add_valid(
    models.IterativeTrimming,
    lambda: {
        "base_gp": surrogates.valid(SingleTaskGPSurrogate).obj(),
        "alpha1": 0.5,
        "alpha2": 0.975,
        "nsh": 2,
        "ncc": 2,
        "nrw": 1,
    },
)



from enum import Enum


class ScalerEnum(Enum):
    """Enumeration class of supported scalers
    Currently, normalization and standardization are implemented.
    """

    NORMALIZE = "NORMALIZE"
    STANDARDIZE = "STANDARDIZE"
    IDENTITY = "IDENTITY"



import random

import bofire.data_models.constraints.api as constraints
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    constraints.LinearEqualityConstraint,
    lambda: {
        "features": ["f1", "f2", "f3"],
        "coefficients": [random.randint(1, 10) for _ in range(3)],
        "rhs": random.random(),
    },
)
specs.add_valid(
    constraints.LinearInequalityConstraint,
    lambda: {
        "features": ["f1", "f2", "f3"],
        "coefficients": [random.randint(1, 10) for _ in range(3)],
        "rhs": random.random(),
    },
)
specs.add_valid(
    constraints.NonlinearEqualityConstraint,
    lambda: {
        "expression": "f1*f2",
        "jacobian_expression": "[f2,f1,0]",
        "features": ["f1", "f2", "f3"],
    },
)
specs.add_valid(
    constraints.NonlinearInequalityConstraint,
    lambda: {
        "expression": "f1*f2",
        "jacobian_expression": "[f2,f1,0]",
        "features": ["f1", "f2", "f3"],
    },
)
specs.add_valid(
    constraints.NChooseKConstraint,
    lambda: {
        "features": ["f1", "f2", "f3"],
        "min_count": 1,
        "max_count": 1,
        "none_also_valid": False,
    },
)



import random

import bofire.data_models.kernels.api as kernels
from tests.bofire.data_models.specs.priors import specs as priors
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    kernels.HammondDistanceKernel,
    lambda: {
        "ard": False,
    },
)
specs.add_valid(
    kernels.LinearKernel,
    lambda: {"variance_prior": priors.valid().obj()},
)
specs.add_valid(
    kernels.MaternKernel,
    lambda: {
        "ard": True,
        "nu": random.random(),
        "lengthscale_prior": priors.valid().obj(),
    },
)
specs.add_valid(
    kernels.RBFKernel,
    lambda: {
        "ard": True,
        "lengthscale_prior": priors.valid().obj(),
    },
)
specs.add_valid(
    kernels.ScaleKernel,
    lambda: {
        "base_kernel": specs.valid(kernels.LinearKernel).obj(),
        "outputscale_prior": priors.valid().obj(),
    },
)
specs.add_valid(
    kernels.AdditiveKernel,
    lambda: {
        "kernels": [
            specs.valid(kernels.LinearKernel).obj(),
            specs.valid(kernels.MaternKernel).obj(),
        ]
    },
)
specs.add_valid(
    kernels.MultiplicativeKernel,
    lambda: {
        "kernels": [
            specs.valid(kernels.LinearKernel).obj(),
            specs.valid(kernels.MaternKernel).obj(),
        ]
    },
)
specs.add_valid(
    kernels.TanimotoKernel,
    lambda: {
        "ard": True,
    },
)



import bofire.data_models.objectives.api as objectives
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    objectives.CloseToTargetObjective,
    lambda: {
        "target_value": 42,
        "exponent": 2,
        "w": 1.0,
    },
)
specs.add_valid(
    objectives.MaximizeObjective,
    lambda: {"w": 1.0, "bounds": (0.1, 0.9)},
)
specs.add_valid(
    objectives.MaximizeSigmoidObjective,
    lambda: {
        "steepness": 0.2,
        "tp": 0.3,
        "w": 1.0,
    },
)
specs.add_valid(
    objectives.MinimizeObjective,
    lambda: {"w": 1.0, "bounds": (0.1, 0.9)},
)

specs.add_invalid(
    objectives.MinimizeObjective,
    lambda: {"w": 1.0, "lower_bound": 0.1, "upper_bound": 0.9},
)

specs.add_valid(
    objectives.MinimizeSigmoidObjective,
    lambda: {
        "steepness": 0.2,
        "tp": 0.3,
        "w": 1.0,
    },
)
specs.add_valid(
    objectives.TargetObjective,
    lambda: {
        "w": 1.0,
        "target_value": 0.4,
        "tolerance": 0.4,
        "steepness": 0.3,
    },
)



import bofire.data_models.strategies.api as strategies
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.acquisition_functions.api import qEI, qLogNEHVI, qPI
from bofire.data_models.domain.api import Domain, Inputs
from bofire.data_models.enum import CategoricalMethodEnum, SamplingMethodEnum
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    DiscreteInput,
)
from tests.bofire.data_models.specs.api import domain
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])


strategy_commons = {
    "num_raw_samples": 1024,
    "num_sobol_samples": 512,
    "num_restarts": 8,
    "descriptor_method": CategoricalMethodEnum.EXHAUSTIVE,
    "categorical_method": CategoricalMethodEnum.EXHAUSTIVE,
    "discrete_method": CategoricalMethodEnum.EXHAUSTIVE,
    "surrogate_specs": None,
    "outlier_detection_specs": None,
    "seed": 42,
    "min_experiments_before_outlier_check": 1,
    "frequency_check": 1,
    "frequency_hyperopt": 0,
    "folds": 5,
}


specs.add_valid(
    strategies.QehviStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        **strategy_commons,
    },
)
specs.add_valid(
    strategies.QnehviStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        **strategy_commons,
        "alpha": 0.4,
    },
)
specs.add_valid(
    strategies.QparegoStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        "acquisition_function": qEI().dict(),
        **strategy_commons,
    },
)
specs.add_valid(
    strategies.MoboStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        "acquisition_function": qLogNEHVI().dict(),
        **strategy_commons,
    },
)
specs.add_valid(
    strategies.SoboStrategy,
    lambda: {
        "domain": Himmelblau().domain.dict(),
        **strategy_commons,
        "acquisition_function": qPI(tau=0.1).dict(),
    },
)
specs.add_valid(
    strategies.AdditiveSoboStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        "acquisition_function": qPI(tau=0.1).dict(),
        "use_output_constraints": True,
        **strategy_commons,
    },
)
specs.add_valid(
    strategies.MultiplicativeSoboStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        **strategy_commons,
        "acquisition_function": qPI(tau=0.1).dict(),
    },
)
specs.add_valid(
    strategies.CustomSoboStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        **strategy_commons,
        "acquisition_function": qPI(tau=0.1).dict(),
        "use_output_constraints": True,
    },
)
specs.add_valid(
    strategies.RandomStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        "seed": 42,
    },
)

specs.add_valid(
    strategies.PolytopeSampler,
    lambda: {
        "domain": Domain(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i}", bounds=(0, 1)) for i in range(2)
                ]
            ),
        ),
        "fallback_sampling_method": SamplingMethodEnum.UNIFORM,
        "seed": 42,
    },
)
specs.add_valid(
    strategies.RejectionSampler,
    lambda: {
        "domain": Domain(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i}", bounds=(0, 1)) for i in range(2)
                ]
            )
        ),
        "max_iters": 1000,
        "num_base_samples": 1000,
        "sampling_method": SamplingMethodEnum.UNIFORM,
        "seed": 42,
    },
)
specs.add_valid(
    strategies.DoEStrategy,
    lambda: {
        "domain": domain.valid().obj().dict(),
        "formula": "linear",
        "optimization_strategy": "default",
        "verbose": False,
        "seed": 42,
    },
)

tempdomain = domain.valid().obj().dict()

specs.add_valid(
    strategies.StepwiseStrategy,
    lambda: {
        "domain": tempdomain,
        "steps": [
            strategies.Step(
                strategy_data=strategies.RandomStrategy(domain=tempdomain),
                condition=strategies.NumberOfExperimentsCondition(n_experiments=10),
                max_parallelism=2,
            ).dict(),
            strategies.Step(
                strategy_data=strategies.QehviStrategy(
                    domain=tempdomain,
                ),
                condition=strategies.NumberOfExperimentsCondition(n_experiments=30),
                max_parallelism=2,
            ).dict(),
        ],
        "seed": 42,
    },
)


specs.add_valid(
    strategies.FactorialStrategy,
    lambda: {
        "domain": Domain(
            inputs=Inputs(
                features=[
                    CategoricalInput(key="alpha", categories=["a", "b", "c"]),
                    DiscreteInput(key="beta", values=[1.0, 2, 3.0, 4.0]),
                ]
            )
        ),
        "seed": 42,
    },
)



from typing import Literal, Union

import numpy as np
import pandas as pd

from bofire.data_models.objectives.objective import (
    ConstrainedObjective,
    Objective,
    TGt0,
    TWeight,
)


class SigmoidObjective(Objective, ConstrainedObjective):
    """Base class for all sigmoid shaped objectives

    Attributes:
        w (float): float between zero and one for weighting the objective.
        steepness (float): Steepness of the sigmoid function. Has to be greater than zero.
        tp (float): Turning point of the sigmoid function.
    """

    steepness: TGt0
    tp: float
    w: TWeight = 1


class MaximizeSigmoidObjective(SigmoidObjective):
    """Class for a maximizing sigmoid objective

    Attributes:
        w (float): float between zero and one for weighting the objective.
        steepness (float): Steepness of the sigmoid function. Has to be greater than zero.
        tp (float): Turning point of the sigmoid function.

    """

    type: Literal["MaximizeSigmoidObjective"] = "MaximizeSigmoidObjective"

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """The call function returning a sigmoid shaped reward for passed x values.

        Args:
            x (np.ndarray): An array of x values

        Returns:
            np.ndarray: A reward calculated with a sigmoid function. The stepness and the tipping point can be modified via passed arguments.
        """
        return 1 / (1 + np.exp(-1 * self.steepness * (x - self.tp)))


class MinimizeSigmoidObjective(SigmoidObjective):
    """Class for a minimizing a sigmoid objective

    Attributes:
        w (float): float between zero and one for weighting the objective.
        steepness (float): Steepness of the sigmoid function. Has to be greater than zero.
        tp (float): Turning point of the sigmoid function.
    """

    type: Literal["MinimizeSigmoidObjective"] = "MinimizeSigmoidObjective"

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """The call function returning a sigmoid shaped reward for passed x values.

        Args:
            x (np.ndarray): An array of x values

        Returns:
            np.ndarray: A reward calculated with a sigmoid function. The stepness and the tipping point can be modified via passed arguments.
        """
        return 1 - 1 / (1 + np.exp(-1 * self.steepness * (x - self.tp)))



from typing import Dict, Type

from bofire.data_models.surrogates import api as data_models
from bofire.surrogates.empirical import EmpiricalSurrogate
from bofire.surrogates.fully_bayesian import SaasSingleTaskGPSurrogate
from bofire.surrogates.mixed_single_task_gp import MixedSingleTaskGPSurrogate
from bofire.surrogates.mlp import MLPEnsemble
from bofire.surrogates.random_forest import RandomForestSurrogate
from bofire.surrogates.single_task_gp import SingleTaskGPSurrogate
from bofire.surrogates.surrogate import Surrogate
from bofire.surrogates.xgb import XGBoostSurrogate

SURROGATE_MAP: Dict[Type[data_models.Surrogate], Type[Surrogate]] = {
    data_models.EmpiricalSurrogate: EmpiricalSurrogate,
    data_models.RandomForestSurrogate: RandomForestSurrogate,
    data_models.SingleTaskGPSurrogate: SingleTaskGPSurrogate,
    data_models.MixedSingleTaskGPSurrogate: MixedSingleTaskGPSurrogate,
    data_models.MLPEnsemble: MLPEnsemble,
    data_models.SaasSingleTaskGPSurrogate: SaasSingleTaskGPSurrogate,
    data_models.XGBoostSurrogate: XGBoostSurrogate,
    data_models.LinearSurrogate: SingleTaskGPSurrogate,
    data_models.PolynomialSurrogate: SingleTaskGPSurrogate,
    data_models.TanimotoGPSurrogate: SingleTaskGPSurrogate,
}


def map(data_model: data_models.Surrogate) -> Surrogate:
    cls = SURROGATE_MAP[data_model.__class__]
    return cls.from_spec(data_model=data_model)



from typing import Literal, Type

from pydantic import confloat

from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
    MinimizeSigmoidObjective,
    Objective,
    TargetObjective,
)
from bofire.data_models.strategies.predictives.qehvi import QehviStrategy


class QnehviStrategy(QehviStrategy):
    type: Literal["QnehviStrategy"] = "QnehviStrategy"
    alpha: confloat(ge=0, le=0.5) = 0.0  # type: ignore

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        """Method to check if a objective type is implemented for the strategy

        Args:
            my_type (Type[Objective]): Objective class

        Returns:
            bool: True if the objective type is valid for the strategy chosen, False otherwise
        """
        return my_type in [
            MaximizeObjective,
            MinimizeObjective,
            MinimizeSigmoidObjective,
            MaximizeSigmoidObjective,
            TargetObjective,
            CloseToTargetObjective,
        ]



from __future__ import annotations

import itertools
import warnings
from enum import Enum
from typing import Dict, List, Literal, Optional, Sequence, Tuple, Type, Union, cast

import numpy as np
import pandas as pd
from pydantic import Field, validate_arguments
from scipy.stats.qmc import LatinHypercube, Sobol

from bofire.data_models.base import BaseModel, filter_by_attribute, filter_by_class
from bofire.data_models.enum import CategoricalEncodingEnum, SamplingMethodEnum
from bofire.data_models.features.api import (
    _CAT_SEP,
    AnyFeature,
    AnyInput,
    AnyOutput,
    CategoricalDescriptorInput,
    CategoricalInput,
    CategoricalMolecularInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Input,
    MolecularInput,
    Output,
    TInputTransformSpecs,
)
from bofire.data_models.molfeatures.api import MolFeatures
from bofire.data_models.objectives.api import AbstractObjective, Objective

FeatureSequence = Union[List[AnyFeature], Tuple[AnyFeature]]


class Features(BaseModel):
    """Container of features, both input and output features are allowed.

    Attributes:
        features (List(Features)): list of the features.
    """

    type: Literal["Features"] = "Features"
    features: FeatureSequence = Field(default_factory=lambda: [])

    def __iter__(self):
        return iter(self.features)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        return self.features[i]

    def __add__(self, other: Union[Sequence[AnyFeature], Features]):
        if isinstance(other, Features):
            other_feature_seq = other.features
        else:
            other_feature_seq = other
        new_feature_seq = list(itertools.chain(self.features, other_feature_seq))

        def is_feats_of_type(feats, ftype_collection, ftype_element):
            return isinstance(feats, ftype_collection) or (
                not isinstance(feats, Features)
                and (len(feats) > 0 and isinstance(feats[0], ftype_element))
            )

        def is_infeats(feats):
            return is_feats_of_type(feats, Inputs, Input)

        def is_outfeats(feats):
            return is_feats_of_type(feats, Outputs, Output)

        if is_infeats(self) and is_infeats(other):
            return Inputs(features=cast(Tuple[AnyInput, ...], new_feature_seq))
        if is_outfeats(self) and is_outfeats(other):
            return Outputs(features=cast(Tuple[AnyOutput, ...], new_feature_seq))
        return Features(features=new_feature_seq)

    def get_by_key(self, key: str) -> AnyFeature:
        """Get a feature by its key.

        Args:
            key (str): Feature key of the feature of interest

        Returns:
            Feature: Feature of interest
        """
        return {f.key: f for f in self.features}[key]

    def get_by_keys(self, keys: Sequence[str]) -> Features:
        """Get features of the domain specified by its keys.

        Args:
            keys (Sequence[str]): List of the keys of the features that should be
                returned.

        Returns:
            Features: Features object with the requested features.
        """
        return self.__class__(features=sorted([self.get_by_key(key) for key in keys]))

    def get(
        self,
        includes: Union[Type, List[Type]] = AnyFeature,
        excludes: Union[Type, List[Type]] = None,
        exact: bool = False,
    ) -> Features:
        """get features of the domain

        Args:
            includes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be returned. Defaults to Feature.
            excludes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact class listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.
            by_attribute (str, optional): If set it is filtered by the attribute specified in by `by_attribute`. Defaults to None.

        Returns:
            List[Feature]: List of features in the domain fitting to the passed requirements.
        """
        return self.__class__(
            features=sorted(
                filter_by_class(
                    self.features,
                    includes=includes,
                    excludes=excludes,
                    exact=exact,
                )
            )
        )

    def get_keys(
        self,
        includes: Union[Type, List[Type]] = AnyFeature,
        excludes: Union[Type, List[Type]] = None,
        exact: bool = False,
    ) -> List[str]:
        """Method to get feature keys of the domain

        Args:
            includes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be returned. Defaults to Feature.
            excludes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact class listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.

        Returns:
            List[str]: List of feature keys fitting to the passed requirements.
        """
        return [
            f.key
            for f in self.get(
                includes=includes,
                excludes=excludes,
                exact=exact,
            )
        ]


class Inputs(Features):
    """Container of input features, only input features are allowed.

    Attributes:
        features (List(Inputs)): list of the features.
    """

    type: Literal["Inputs"] = "Inputs"
    features: Sequence[AnyInput] = Field(default_factory=lambda: [])

    def get_fixed(self) -> "Inputs":
        """Gets all features in `self` that are fixed and returns them as new `Inputs` object.

        Returns:
            Inputs: Input features object containing only fixed features.
        """
        return Inputs(features=[feat for feat in self if feat.is_fixed()])  # type: ignore

    def get_free(self) -> "Inputs":
        """Gets all features in `self` that are not fixed and returns them as new `Inputs` object.

        Returns:
            Inputs: Input features object containing only non-fixed features.
        """
        return Inputs(features=[feat for feat in self if not feat.is_fixed()])  # type: ignore

    @validate_arguments
    def sample(
        self,
        n: int = 1,
        method: SamplingMethodEnum = SamplingMethodEnum.UNIFORM,
    ) -> pd.DataFrame:
        """Draw sobol samples

        Args:
            n (int, optional): Number of samples, has to be larger than 0. Defaults to 1.
            method (SamplingMethodEnum, optional): Method to use, implemented methods are `UNIFORM`, `SOBOL` and `LHS`.
                Defaults to `UNIFORM`.

        Returns:
            pd.DataFrame: Dataframe containing the samples.
        """
        if method == SamplingMethodEnum.UNIFORM:
            return self.validate_candidates(
                pd.concat([feat.sample(n) for feat in self.get(Input)], axis=1)  # type: ignore
            )
        free_features = self.get_free()
        if method == SamplingMethodEnum.SOBOL:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                X = Sobol(len(free_features)).random(n)
        else:
            X = LatinHypercube(len(free_features)).random(n)
        res = []
        for i, feat in enumerate(free_features):
            if isinstance(feat, ContinuousInput):
                x = feat.from_unit_range(X[:, i])
            elif isinstance(feat, (DiscreteInput, CategoricalInput)):
                if isinstance(feat, DiscreteInput):
                    levels = feat.values
                else:
                    levels = feat.get_allowed_categories()
                bins = np.linspace(0, 1, len(levels) + 1)
                idx = np.digitize(X[:, i], bins) - 1
                x = np.array(levels)[idx]
            else:
                raise (
                    ValueError(
                        f"Unknown input feature with key {feat.key} of type {feat.type}"
                    )
                )
            res.append(pd.Series(x, name=feat.key))
        samples = pd.concat(res, axis=1)
        for feat in self.get_fixed():
            samples[feat.key] = feat.fixed_value()[0]  # type: ignore
        return self.validate_candidates(samples)[self.get_keys(Input)]

    def validate_candidates(self, candidates: pd.DataFrame) -> pd.DataFrame:
        """Validate a pandas dataframe with input feature values.

        Args:
            candidates (pd.Dataframe): Inputs to validate.

        Raises:
            ValueError: Raises a Valueerror if a feature based validation raises an exception.

        Returns:
            pd.Dataframe: Validated dataframe
        """
        for feature in self:
            if feature.key not in candidates:
                raise ValueError(f"no col for input feature `{feature.key}`")
            candidates[feature.key] = feature.validate_candidental(candidates[feature.key])  # type: ignore
        if candidates[self.get_keys()].isnull().to_numpy().any():
            raise ValueError("there are null values")
        if candidates[self.get_keys()].isna().to_numpy().any():
            raise ValueError("there are na values")
        return candidates

    def validate_experiments(
        self, experiments: pd.DataFrame, strict=False
    ) -> pd.DataFrame:
        for feature in self:
            if feature.key not in experiments:
                raise ValueError(f"no col for input feature `{feature.key}`")
            experiments[feature.key] = feature.validate_experimental(experiments[feature.key], strict=strict)  # type: ignore
        if experiments[self.get_keys()].isnull().to_numpy().any():
            raise ValueError("there are null values")
        if experiments[self.get_keys()].isna().to_numpy().any():
            raise ValueError("there are na values")
        return experiments

    def get_categorical_combinations(
        self,
        include: Union[Type, List[Type]] = Input,
        exclude: Union[Type, List[Type]] = None,
    ):
        """get a list of tuples pairing the feature keys with a list of valid categories

        Args:
            include (Feature, optional): Features to be included. Defaults to Input.
            exclude (Feature, optional): Features to be excluded, e.g. subclasses of the included features. Defaults to None.

        Returns:
            List[(str, List[str])]: Returns a list of tuples pairing the feature keys with a list of valid categories (str)
        """
        features = [
            f
            for f in self.get(includes=include, excludes=exclude)
            if (isinstance(f, CategoricalInput) and not f.is_fixed())
        ]
        list_of_lists = [
            [(f.key, cat) for cat in f.get_allowed_categories()] for f in features
        ]

        discretes = [
            f
            for f in self.get(includes=include, excludes=exclude)
            if (isinstance(f, DiscreteInput) and not f.is_fixed())
        ]

        list_of_lists_2 = [[(d.key, v) for v in d.values] for d in discretes]

        list_of_lists = list_of_lists + list_of_lists_2

        return list(itertools.product(*list_of_lists))

    # transformation related methods
    def _get_transform_info(
        self, specs: TInputTransformSpecs
    ) -> Tuple[Dict[str, Tuple[int]], Dict[str, Tuple[str]]]:
        """Generates two dictionaries. The first one specifies which key is mapped to
        which column indices when applying `transform`. The second one specifies
        which key is mapped to which transformed keys.

        Args:
            specs (TInputTransformSpecs): Dictionary specifying which
                input feature is transformed by which encoder.

        Returns:
            Dict[str, Tuple[int]]: Dictionary mapping feature keys to column indices.
            Dict[str, Tuple[str]]: Dictionary mapping feature keys to transformed feature
                keys.
        """
        self._validate_transform_specs(specs)
        features2idx = {}
        features2names = {}
        counter = 0
        for _, feat in enumerate(self.get()):
            if feat.key not in specs.keys():
                features2idx[feat.key] = (counter,)
                features2names[feat.key] = (feat.key,)
                counter += 1
            elif specs[feat.key] == CategoricalEncodingEnum.ONE_HOT:
                assert isinstance(feat, CategoricalInput)
                features2idx[feat.key] = tuple(
                    (np.array(range(len(feat.categories))) + counter).tolist()
                )
                features2names[feat.key] = tuple(
                    [f"{feat.key}{_CAT_SEP}{c}" for c in feat.categories]
                )
                counter += len(feat.categories)
            elif specs[feat.key] == CategoricalEncodingEnum.ORDINAL:
                features2idx[feat.key] = (counter,)
                features2names[feat.key] = (feat.key,)
                counter += 1
            elif specs[feat.key] == CategoricalEncodingEnum.DUMMY:
                assert isinstance(feat, CategoricalInput)
                features2idx[feat.key] = tuple(
                    (np.array(range(len(feat.categories) - 1)) + counter).tolist()
                )
                features2names[feat.key] = tuple(
                    [f"{feat.key}{_CAT_SEP}{c}" for c in feat.categories[1:]]
                )
                counter += len(feat.categories) - 1
            elif specs[feat.key] == CategoricalEncodingEnum.DESCRIPTOR:
                assert isinstance(feat, CategoricalDescriptorInput)
                features2idx[feat.key] = tuple(
                    (np.array(range(len(feat.descriptors))) + counter).tolist()
                )
                features2names[feat.key] = tuple(
                    [f"{feat.key}{_CAT_SEP}{d}" for d in feat.descriptors]
                )
                counter += len(feat.descriptors)
            elif isinstance(specs[feat.key], MolFeatures):
                assert isinstance(feat, MolecularInput)
                descriptor_names = specs[
                    feat.key
                ].get_descriptor_names()  # type: ignore
                features2idx[feat.key] = tuple(
                    (np.array(range(len(descriptor_names))) + counter).tolist()
                )
                features2names[feat.key] = tuple(
                    [f"{feat.key}{_CAT_SEP}{d}" for d in descriptor_names]
                )
                counter += len(descriptor_names)
        return features2idx, features2names

    def transform(
        self, experiments: pd.DataFrame, specs: TInputTransformSpecs
    ) -> pd.DataFrame:
        """Transform a dataframe to the represenation specified in `specs`.

        Currently only input categoricals are supported.

        Args:
            experiments (pd.DataFrame): Data dataframe to be transformed.
            specs (TInputTransformSpecs): Dictionary specifying which
                input feature is transformed by which encoder.

        Returns:
            pd.DataFrame: Transformed dataframe. Only input features are included.
        """
        # TODO: clean this up and move it into the individual classes
        specs = self._validate_transform_specs(specs)
        transformed = []
        for feat in self.get():
            s = experiments[feat.key]
            if feat.key not in specs.keys():
                transformed.append(s)
            elif specs[feat.key] == CategoricalEncodingEnum.ONE_HOT:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.to_onehot_encoding(s))
            elif specs[feat.key] == CategoricalEncodingEnum.ORDINAL:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.to_ordinal_encoding(s))
            elif specs[feat.key] == CategoricalEncodingEnum.DUMMY:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.to_dummy_encoding(s))
            elif specs[feat.key] == CategoricalEncodingEnum.DESCRIPTOR:
                assert isinstance(feat, CategoricalDescriptorInput)
                transformed.append(feat.to_descriptor_encoding(s))
            elif isinstance(specs[feat.key], MolFeatures):
                assert isinstance(feat, MolecularInput)
                transformed.append(feat.to_descriptor_encoding(specs[feat.key], s))  # type: ignore
        return pd.concat(transformed, axis=1)

    def inverse_transform(
        self, experiments: pd.DataFrame, specs: TInputTransformSpecs
    ) -> pd.DataFrame:
        """Transform a dataframe back to the original representations.

        The original applied transformation has to be provided via the specs dictionary.
        Currently only input categoricals are supported.

        Args:
            experiments (pd.DataFrame): Transformed data dataframe.
            specs (TInputTransformSpecs): Dictionary specifying which
                input feature is transformed by which encoder.

        Returns:
            pd.DataFrame: Back transformed dataframe. Only input features are included.
        """
        # TODO: clean this up and move it into the individual classes
        self._validate_transform_specs(specs=specs)
        transformed = []
        for feat in self.get():
            if isinstance(feat, DiscreteInput):
                transformed.append(feat.from_continuous(experiments))
            elif feat.key not in specs.keys():
                transformed.append(experiments[feat.key])
            elif specs[feat.key] == CategoricalEncodingEnum.ONE_HOT:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.from_onehot_encoding(experiments))
            elif specs[feat.key] == CategoricalEncodingEnum.ORDINAL:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.from_ordinal_encoding(experiments[feat.key]))
            elif specs[feat.key] == CategoricalEncodingEnum.DUMMY:
                assert isinstance(feat, CategoricalInput)
                transformed.append(feat.from_dummy_encoding(experiments))
            elif specs[feat.key] == CategoricalEncodingEnum.DESCRIPTOR:
                assert isinstance(feat, CategoricalDescriptorInput)
                transformed.append(feat.from_descriptor_encoding(experiments))
            elif isinstance(specs[feat.key], MolFeatures):
                assert isinstance(feat, CategoricalMolecularInput)
                transformed.append(feat.from_descriptor_encoding(specs[feat.key], experiments))  # type: ignore

        return pd.concat(transformed, axis=1)

    def _validate_transform_specs(
        self, specs: TInputTransformSpecs
    ) -> TInputTransformSpecs:
        """Checks the validity of the transform specs .

        Args:
            specs (TInputTransformSpecs): Transform specs to be validated.
        """
        # first check that the keys in the specs dict are correct also correct feature keys
        # next check that all values are of type CategoricalEncodingEnum or MolFeatures
        for key, value in specs.items():
            try:
                feat = self.get_by_key(key)
            except KeyError:
                raise ValueError(
                    f"Unknown feature with key {key} specified in transform specs."
                )
            # TODO
            # this is ugly, on the long run we have to get rid of the transform enums
            # and replace them with classes, then the following lines collapse into just two
            assert isinstance(feat, Input)
            enums = [t for t in feat.valid_transform_types() if isinstance(t, Enum)]
            no_enums = [
                t for t in feat.valid_transform_types() if not isinstance(t, Enum)
            ]
            if isinstance(value, Enum):
                if value not in enums:
                    raise ValueError(
                        f"Forbidden transform type for feature with key {key}"
                    )
            else:
                if len(no_enums) == 0:
                    raise ValueError(
                        f"Forbidden transform type for feature with key {key}"
                    )
                if not isinstance(value, tuple(no_enums)):  # type: ignore
                    raise ValueError(
                        f"Forbidden transform type for feature with key {key}"
                    )
        return specs

    def get_bounds(
        self,
        specs: TInputTransformSpecs,
        experiments: Optional[pd.DataFrame] = None,
    ) -> Tuple[List[float], List[float]]:
        """Returns the boundaries of the optimization problem based on the transformations
        defined in the  `specs` dictionary.

        Args:
            specs (TInputTransformSpecs): Dictionary specifying which
                input feature is transformed by which encoder.
            experiments (Optional[pd.DataFrame], optional): Dataframe with input features.
                If provided the real feature bounds are returned based on both the opt.
                feature bounds and the extreme points in the dataframe. Defaults to None,

        Raises:
            ValueError: If a feature type is not known.
            ValueError: If no transformation is provided for a categorical feature.

        Returns:
            Tuple[List[float], List[float]]: list with lower bounds, list with upper bounds.
        """
        self._validate_transform_specs(specs=specs)

        lower = []
        upper = []

        for feat in self.get():
            lo, up = feat.get_bounds(  # type: ignore
                transform_type=specs.get(feat.key),  # type: ignore
                values=experiments[feat.key] if experiments is not None else None,  # type: ignore
            )
            lower += lo
            upper += up
        return lower, upper


class Outputs(Features):
    """Container of output features, only output features are allowed.

    Attributes:
        features (List(Outputs)): list of the features.
    """

    type: Literal["Outputs"] = "Outputs"
    features: Sequence[AnyOutput] = Field(default_factory=lambda: [])

    def get_by_objective(
        self,
        includes: Union[
            List[Type[AbstractObjective]],
            Type[AbstractObjective],
            Type[Objective],
        ] = Objective,
        excludes: Union[
            List[Type[AbstractObjective]],
            Type[AbstractObjective],
            None,
        ] = None,
        exact: bool = False,
    ) -> "Outputs":
        """Get output features filtered by the type of the attached objective.

        Args:
            includes (Union[List[TObjective], TObjective], optional): Objective class or list of objective classes
                to be returned. Defaults to Objective.
            excludes (Union[List[TObjective], TObjective, None], optional): Objective class or list of specific objective classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact classes listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.

        Returns:
            List[AnyOutput]: List of output features fitting to the passed requirements.
        """
        if len(self.features) == 0:
            return Outputs(features=[])
        else:
            return Outputs(
                features=sorted(
                    filter_by_attribute(
                        self.get(ContinuousOutput).features,
                        lambda of: of.objective,
                        includes,
                        excludes,
                        exact,
                    )
                )
            )

    def get_keys_by_objective(
        self,
        includes: Union[
            List[Type[AbstractObjective]],
            Type[AbstractObjective],
            Type[Objective],
        ] = Objective,
        excludes: Union[
            List[Type[AbstractObjective]], Type[AbstractObjective], None
        ] = None,
        exact: bool = False,
    ) -> List[str]:
        """Get keys of output features filtered by the type of the attached objective.

        Args:
            includes (Union[List[TObjective], TObjective], optional): Objective class or list of objective classes
                to be returned. Defaults to Objective.
            excludes (Union[List[TObjective], TObjective, None], optional): Objective class or list of specific objective classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact classes listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.

        Returns:
            List[str]: List of output feature keys fitting to the passed requirements.
        """
        return [f.key for f in self.get_by_objective(includes, excludes, exact)]

    def __call__(
        self, experiments: pd.DataFrame, predictions: bool = False
    ) -> pd.DataFrame:
        """Evaluate the objective for every feature.

        Args:
            experiments (pd.DataFrame): Experiments for which the objectives should be evaluated.
            predictions (bool, optional): If True use the prediction columns in the dataframe to calc the
                desirabilities `f"{feat.key}_pred`.

        Returns:
            pd.DataFrame: Objective values for the experiments of interest.
        """
        desis = pd.concat(
            [
                feat(experiments[f"{feat.key}_pred" if predictions else feat.key])  # type: ignore
                for feat in self.features
                if feat.objective is not None
            ],
            axis=1,
        )
        return desis.rename(
            {
                f"{feat.key}_pred" if predictions else feat.key: f"{feat.key}_des"
                for feat in self.features
                if feat.objective is not None
            },
            axis=1,
        )

    def add_valid_columns(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Add the `valid_{feature.key}` columns to the experiments dataframe,
        in case that they are not present.

        Args:
            experiments (pd.DataFrame): Dataframe holding the experiments.

        Returns:
            pd.DataFrame: Dataframe holding the experiments.
        """
        valid_keys = [
            f"valid_{output_feature_key}" for output_feature_key in self.get_keys()
        ]
        for valid_key in valid_keys:
            if valid_key not in experiments:
                experiments[valid_key] = True
            else:
                try:
                    experiments[valid_key] = (
                        experiments[valid_key].astype(int).astype(bool)
                    )
                except ValueError:
                    raise ValueError(f"Column {valid_key} cannot casted to dtype bool.")
        return experiments

    def validate_experiments(self, experiments: pd.DataFrame) -> pd.DataFrame:
        for feat in self.get():
            if feat.key not in experiments:
                raise ValueError(f"no col for input feature `{feat.key}`")
            experiments[feat.key] = feat.validate_experimental(experiments[feat.key])
        experiments = self.add_valid_columns(experiments=experiments)
        return experiments

    def validate_candidates(self, candidates: pd.DataFrame) -> pd.DataFrame:
        # for each continuous output feature with an attached objective object
        # ToDo: adjust it for the CategoricalOutput
        cols = list(
            itertools.chain.from_iterable(
                [
                    [f"{key}_pred", f"{key}_sd", f"{key}_des"]
                    for key in self.get_keys_by_objective(Objective)
                ]
                + [
                    [f"{key}_pred", f"{key}_sd"]
                    for key in self.get_keys_by_objective(
                        excludes=Objective, includes=None  # type: ignore
                    )
                ]
            )
        )
        # check that pred, sd, and des cols are specified and numerical
        for col in cols:
            if col not in candidates:
                raise ValueError(f"missing column {col}")
            try:
                candidates[col] = pd.to_numeric(candidates[col], errors="raise").astype(
                    "float64"
                )
            except ValueError:
                raise ValueError(f"Not all values of column `{col}` are numerical.")
            if candidates[col].isnull().to_numpy().any():
                raise ValueError(f"Nan values are present in {col}.")
        return candidates

    def preprocess_experiments_one_valid_output(
        self,
        output_feature_key: str,
        experiments: pd.DataFrame,
    ) -> pd.DataFrame:
        """Method to get a dataframe where non-valid entries of the provided output feature are removed

        Args:
            experiments (pd.DataFrame): Dataframe with experimental data
            output_feature_key (str): The feature based on which non-valid entries rows are removed

        Returns:
            pd.DataFrame: Dataframe with all experiments where only valid entries of the specific feature are included
        """
        clean_exp = experiments.loc[
            (experiments["valid_%s" % output_feature_key] == 1)
            & (experiments[output_feature_key].notna())
        ]

        return clean_exp

    def preprocess_experiments_all_valid_outputs(
        self,
        experiments: pd.DataFrame,
        output_feature_keys: Optional[List] = None,
    ) -> pd.DataFrame:
        """Method to get a dataframe where non-valid entries of all output feature are removed

        Args:
            experiments (pd.DataFrame): Dataframe with experimental data
            output_feature_keys (Optional[List], optional): List of output feature keys which should be considered for removal of invalid values. Defaults to None.

        Returns:
            pd.DataFrame: Dataframe with all experiments where only valid entries of the selected features are included
        """
        if (output_feature_keys is None) or (len(output_feature_keys) == 0):
            output_feature_keys = self.get_keys(Output)

        clean_exp = experiments.query(
            " & ".join(["(`valid_%s` > 0)" % key for key in output_feature_keys])
        )
        clean_exp = clean_exp.dropna(subset=output_feature_keys)

        return clean_exp

    def preprocess_experiments_any_valid_output(
        self, experiments: pd.DataFrame
    ) -> pd.DataFrame:
        """Method to get a dataframe where at least one output feature has a valid entry

        Args:
            experiments (pd.DataFrame): Dataframe with experimental data

        Returns:
            pd.DataFrame: Dataframe with all experiments where at least one output feature has a valid entry
        """

        output_feature_keys = self.get_keys(Output)

        # clean_exp = experiments.query(" or ".join(["(valid_%s > 0)" % key for key in output_feature_keys]))
        # clean_exp = clean_exp.query(" or ".join(["%s.notna()" % key for key in output_feature_keys]))

        assert experiments is not None
        clean_exp = experiments.query(
            " or ".join(
                [
                    "((`valid_%s` >0) & `%s`.notna())" % (key, key)
                    for key in output_feature_keys
                ]
            )
        )
        return clean_exp



from typing import Union

from bofire.data_models.surrogates.scaler import ScalerEnum

try:
    from bofire.data_models.surrogates.botorch import BotorchSurrogate
    from bofire.data_models.surrogates.botorch_surrogates import (
        AnyBotorchSurrogate,
        BotorchSurrogates,
    )
    from bofire.data_models.surrogates.empirical import EmpiricalSurrogate
    from bofire.data_models.surrogates.fully_bayesian import SaasSingleTaskGPSurrogate
    from bofire.data_models.surrogates.linear import LinearSurrogate
    from bofire.data_models.surrogates.mixed_single_task_gp import (
        MixedSingleTaskGPSurrogate,
    )
    from bofire.data_models.surrogates.mlp import MLPEnsemble
    from bofire.data_models.surrogates.polynomial import PolynomialSurrogate
    from bofire.data_models.surrogates.random_forest import RandomForestSurrogate
    from bofire.data_models.surrogates.single_task_gp import (
        SingleTaskGPHyperconfig,
        SingleTaskGPSurrogate,
    )
    from bofire.data_models.surrogates.surrogate import Surrogate
    from bofire.data_models.surrogates.tanimoto_gp import TanimotoGPSurrogate
    from bofire.data_models.surrogates.trainable import MeanAggregation, SumAggregation
    from bofire.data_models.surrogates.xgb import XGBoostSurrogate

    AbstractSurrogate = Union[Surrogate, BotorchSurrogate, EmpiricalSurrogate]

    AnySurrogate = Union[
        EmpiricalSurrogate,
        RandomForestSurrogate,
        SingleTaskGPSurrogate,
        MixedSingleTaskGPSurrogate,
        MLPEnsemble,
        SaasSingleTaskGPSurrogate,
        XGBoostSurrogate,
        LinearSurrogate,
        PolynomialSurrogate,
        TanimotoGPSurrogate,
    ]

    AnyTrainableSurrogate = Union[
        RandomForestSurrogate,
        SingleTaskGPSurrogate,
        MixedSingleTaskGPSurrogate,
        MLPEnsemble,
        SaasSingleTaskGPSurrogate,
        XGBoostSurrogate,
        LinearSurrogate,
        PolynomialSurrogate,
        TanimotoGPSurrogate,
    ]
except ImportError:
    # with the minimal installationwe don't have botorch
    pass



import warnings
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import Field, root_validator, validator
from scipy.integrate import simps
from scipy.stats import fisher_exact, kendalltau, norm, pearsonr, spearmanr
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
)

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.domain import is_numeric
from bofire.data_models.enum import RegressionMetricsEnum, UQRegressionMetricsEnum
from bofire.data_models.validated import ValidatedDataFrame, ValidatedSeries


def _mean_absolute_error(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the mean absolute error.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: mean absolute error
    """
    return mean_absolute_error(observed, predicted)


def _mean_squared_error(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the mean squared error.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: mean squared error
    """
    return mean_squared_error(observed, predicted)


def _mean_absolute_percentage_error(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the mean percentage error.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: mean percentage error
    """
    return mean_absolute_percentage_error(observed, predicted)


def _r2_score(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the R2 score.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: R2 score.
    """
    return float(r2_score(observed, predicted))


def _pearson(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the Pearson correlation coefficient.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: Pearson correlation coefficient.
    """
    with np.errstate(invalid="ignore"):
        rho, _ = pearsonr(predicted, observed)
    return float(rho)


def _spearman(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the Spearman correlation coefficient.

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: Spearman correlation coefficient.
    """
    with np.errstate(invalid="ignore"):
        rho, _ = spearmanr(predicted, observed)
    return float(rho)


def _fisher_exact_test_p(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Test if the model is able to distuinguish the bottom half of the observations from the top half.

    For this purpose Fisher's excat test is used together with the observations and predictions. The
    p value is returned. A low p value indicates that the model has some ability to distuiguish high from
    low values. A high p value indcates that the model cannot identify the difference or that the
    observations are too noisy to be able to tell.

    This implementation is taken from Ax: https://github.com/facebook/Ax/blob/main/ax/modelbridge/cross_validation.py

    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
            Ignored in the calculation. Defaults to None.

    Returns:
        float: p value of the test.
    """
    n_half = len(observed) // 2
    top_obs = observed.argsort(axis=0)[-n_half:]
    top_est = predicted.argsort(axis=0)[-n_half:]
    # Construct contingency table
    tp = len(set(top_est).intersection(top_obs))
    fp = n_half - tp
    fn = n_half - tp
    tn = (len(observed) - n_half) - (n_half - tp)
    table = np.array([[tp, fp], [fn, tn]])
    # Compute the test statistic
    _, p = fisher_exact(table, alternative="greater")
    return float(p)


def _spearman_UQ(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the Spearman correlation coefficient between the models absolute error
    and the uncertainty - non-linear correlation.

    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.

    Returns:
        float: Spearman correlation coefficient.
    """
    if standard_deviation is None:
        warnings.warn(
            "Uncertainty quantification without standard deviation is not possible",
            UserWarning,
        )
        return np.nan
    else:
        ae = np.abs(observed - predicted)  # type: ignore
        with np.errstate(invalid="ignore"):
            rho, _ = spearmanr(ae, standard_deviation)
        return float(rho)


def _pearson_UQ(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the Pearson correlation coefficient between the models absolute error
    and the uncertainty - linear correlation.

    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.

    Returns:
        float: Pearson correlation coefficient.
    """
    if standard_deviation is None:
        warnings.warn(
            "Uncertainty quantification without standard deviation is not possible",
            UserWarning,
        )
        return np.nan
    else:
        ae = np.abs(observed - predicted)  # type: ignore
        with np.errstate(invalid="ignore"):
            rho, _ = pearsonr(ae, standard_deviation)
        return float(rho)


def _kendall_UQ(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
) -> float:
    """Calculates the Kendall correlation coefficient between the models absolute error
    and the uncertainty - linear correlation.

    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.

    Returns:
        float: Kendall correlation coefficient.
    """
    if standard_deviation is None:
        warnings.warn(
            "Uncertainty quantification without standard deviation is not possible",
            UserWarning,
        )
        return np.nan
    else:
        ae = np.abs(observed - predicted)  # type: ignore
        with np.errstate(invalid="ignore"):
            rho, _ = kendalltau(ae, standard_deviation)
        return float(rho)


def _CVPPDiagram(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
    num_bins: int = 10,
) -> Tuple[np.ndarray, np.ndarray]:
    """Metric introduced in arXiv:2010.01118 [cs.LG] based on cross-validatory
    predictive p-values.
    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
        num_bins (int): number of bins for getting quantiles.

    Returns:
        np.ndarray: quantiles.
        np.ndarray: callibration score for each quantile.
    """
    if standard_deviation is None:
        raise ValueError(
            "Calibration metric without standard deviation is not possible"
        )
    else:
        lhs = np.abs((predicted - observed) / standard_deviation)  # type: ignore
        qs = np.linspace(0, 1, num_bins)
        Cqs = np.empty(qs.shape)
        for ix, q in enumerate(qs):
            rhs = norm.ppf(((1.0 + q) / 2.0), loc=0.0, scale=1.0)
            Cqs[ix] = np.sum((lhs < rhs).astype(int)) / observed.shape[0]

        return qs, Cqs


def _MaximumMiscalibration(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
    num_bins: int = 10,
) -> float:
    """Miscalibration metric with CVPP
    WARNING - this metric only diagnoses systematic over- or under-
    confidence, i.e. a model that is overconfident for ~half of the
    quantiles and under-confident for ~half will still have a MiscalibrationArea
    of ~0.
    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
        num_bins (int): number of bins for getting quantiles.

    Returns:
        float: maximum miscalibration
    """
    try:
        qs, Cqs = _CVPPDiagram(
            observed=observed,
            predicted=predicted,
            standard_deviation=standard_deviation,
            num_bins=num_bins,
        )
        res = np.max(np.abs(Cqs - qs))  # type: ignore

        return float(res)
    except ValueError:
        warnings.warn(
            "Calibration metric without standard deviation is not possible", UserWarning
        )
        return np.nan


def _MiscalibrationArea(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
    num_bins: int = 10,
) -> float:
    """Miscalibration area metric with CVPP
    WARNING - this metric only diagnoses systematic over- or under-
    confidence, i.e. a model that is overconfident for ~half of the
    quantiles and under-confident for ~half will still have a MiscalibrationArea
    of ~0.
    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
        num_bins (int): number of bins for getting quantiles.

    Returns:
        float: total miscalibration area
    """
    try:
        qs, Cqs = _CVPPDiagram(
            observed=observed,
            predicted=predicted,
            standard_deviation=standard_deviation,
            num_bins=num_bins,
        )
        res = simps(Cqs - qs, qs)  # type: ignore

        return float(res)
    except ValueError:
        warnings.warn(
            "Calibration metric without standard deviation is not possible", UserWarning
        )
        return np.nan


def _AbsoluteMiscalibrationArea(
    observed: np.ndarray,
    predicted: np.ndarray,
    standard_deviation: Optional[np.ndarray] = None,
    num_bins: int = 10,
) -> float:
    """absolute miscalibration area metric with CVPP
    This implementation is taken from : https://github.com/aspuru-guzik-group/dionysus/blob/main/dionysus/uncertainty_metrics.py


    Args:
        observed (np.ndarray): Observed data.
        predicted (np.ndarray): Predicted data.
        standard_deviation (Optional[np.ndarray], optional): Predicted standard deviation.
        num_bins (int): number of bins for getting quantiles.

    Returns:
        float: absolute miscalibration area
    """
    try:
        qs, Cqs = _CVPPDiagram(
            observed=observed,
            predicted=predicted,
            standard_deviation=standard_deviation,
            num_bins=num_bins,
        )
        res = simps(np.abs(Cqs - qs), qs)  # type: ignore

        return float(res)
    except ValueError:
        warnings.warn(
            "Calibration metric without standard deviation is not possible", UserWarning
        )
        return np.nan


metrics = {
    RegressionMetricsEnum.MAE: _mean_absolute_error,
    RegressionMetricsEnum.MSD: _mean_squared_error,
    RegressionMetricsEnum.R2: _r2_score,
    RegressionMetricsEnum.MAPE: _mean_absolute_percentage_error,
    RegressionMetricsEnum.PEARSON: _pearson,
    RegressionMetricsEnum.SPEARMAN: _spearman,
    RegressionMetricsEnum.FISHER: _fisher_exact_test_p,
}

UQ_metrics = {
    UQRegressionMetricsEnum.PEARSON_UQ: _pearson_UQ,
    UQRegressionMetricsEnum.SPEARMAN_UQ: _spearman_UQ,
    UQRegressionMetricsEnum.KENDALL_UQ: _kendall_UQ,
    UQRegressionMetricsEnum.MAXIMUMCALIBRATION: _MaximumMiscalibration,
    UQRegressionMetricsEnum.MISCALIBRATIONAREA: _MiscalibrationArea,
    UQRegressionMetricsEnum.ABSOLUTEMISCALIBRATIONAREA: _AbsoluteMiscalibrationArea,
}

all_metrics = {**metrics, **UQ_metrics}


class CvResult(BaseModel):
    """Container representing the results of one CV fold.

    Attributes:
        key (str): Key of the validated output feature.
        observed (pd.Series): Series holding the observed values
        predicted (pd.Series): Series holding the predicted values
        standard_deviation (pd.Series, optional): Series holding the standard deviation associated with
            the prediction. Defaults to None.
    """

    key: str
    observed: ValidatedSeries
    predicted: ValidatedSeries
    standard_deviation: Optional[ValidatedSeries] = None
    labcodes: Optional[ValidatedSeries] = None
    X: Optional[ValidatedDataFrame] = None

    @root_validator(pre=True)
    def validate_shapes(cls, values):
        if not len(values["predicted"]) == len(values["observed"]):
            raise ValueError(
                f"Predicted values has length {len(values['predicted'])} whereas observed has length {len(values['observed'])}"
            )
        if "standard_deviation" in values and values["standard_deviation"] is not None:
            if not len(values["predicted"]) == len(values["standard_deviation"]):
                raise ValueError(
                    f"Predicted values has length {len(values['predicted'])} whereas standard_deviation has length {len(values['standard_deviation'])}"
                )
        if "labcodes" in values and values["labcodes"] is not None:
            if not len(values["predicted"]) == len(values["labcodes"]):
                raise ValueError(
                    f"Predicted values has length {len(values['predicted'])} whereas labcodes has length {len(values['labcodes'])}"
                )
        if "X" in values and values["X"] is not None:
            if not len(values["predicted"]) == len(values["X"]):
                raise ValueError(
                    f"Predicted values has length {len(values['predicted'])} whereas X has length {len(values['X'])}"
                )
        return values

    @validator("observed")
    def validate_observed(cls, v, values):
        if not is_numeric(v):
            raise ValueError("Not all values of observed are numerical")
        return v

    @validator("predicted")
    def validate_predicted(cls, v, values):
        if not is_numeric(v):
            raise ValueError("Not all values of predicted are numerical")
        return v

    @validator("standard_deviation")
    def validate_standard_deviation(cls, v, values):
        if v is not None:
            if not is_numeric(v):
                raise ValueError("Not all values of standard_deviation are numerical")
        return v

    @property
    def n_samples(self) -> int:
        """Returns the number of samples in the fold.

        Returns:
            int: Number of samples in the split.
        """
        return len(self.observed)

    def get_metric(
        self, metric: Union[RegressionMetricsEnum, UQRegressionMetricsEnum]
    ) -> float:
        """Calculates a metric for the fold.

        Args:
            metric (RegressionMetricsEnum): Metric to calculate.

        Returns:
            float: Metric value.
        """
        if self.n_samples == 1:
            warnings.warn(
                "Metric cannot be calculated for only one sample. Null value will be returned"
            )
            return np.nan
        return all_metrics[metric](self.observed.values, self.predicted.values, self.standard_deviation)  # type: ignore


class CvResults(BaseModel):
    """Container holding all cv folds of a cross-validation run.

    Attributes:
        results (Sequence[CvResult]: Sequence of `CvResult` objects.
    """

    results: Sequence[CvResult]

    @validator("results")
    def validate_results(cls, v, values):
        if len(v) <= 1:
            raise ValueError("`results` sequence has to contain at least two elements.")
        key = v[0].key
        for i in v:
            if i.key != key:
                raise ValueError("`CvResult` objects do not match.")
        for field in ["standard_deviation", "labcodes", "X"]:
            has_field = getattr(v[0], field) is not None
            for i in v:
                has_i = getattr(i, field) is not None
                if has_field != has_i:
                    raise ValueError(
                        f"Either all or none `CvResult` objects contain {field}."
                    )
        # check columns of X
        if v[0].X is not None:
            cols = sorted(v[0].X.columns)
            for i in v:
                if sorted(i.X.columns) != cols:
                    raise ValueError("Columns of X do not match.")
        return v

    def __len__(self) -> int:
        return len(self.results)

    def __iter__(self):
        return iter(self.results)

    def __getitem__(self, i) -> CvResult:
        return self.results[i]

    @property
    def key(self) -> str:
        """Returns name of the feature for which the cross validation was performed.

        Returns:
            str: feature name.
        """
        return self.results[0].key

    @property
    def is_loo(self) -> bool:
        """Checks if the object represents a LOO-CV

        Returns:
            bool: True if LOO-CV else False.
        """
        return (np.array([r.n_samples for r in self.results]) == 1).all()

    def _combine_folds(self) -> CvResult:
        """Combines the `CvResult` splits into one flat array for predicted, observed and standard_deviation.

        Returns:
            Tuple[np.ndarray, np.ndarray, Union[np.ndarray, None]]: One pd.Series for CvResult property.
        """
        observed = pd.concat([cv.observed for cv in self.results], ignore_index=True)
        predicted = pd.concat([cv.predicted for cv in self.results], ignore_index=True)
        if self.results[0].standard_deviation is not None:
            sd = pd.concat([cv.standard_deviation for cv in self.results], ignore_index=True)  # type: ignore
        else:
            sd = None
        if self.results[0].labcodes is not None:
            labcodes = pd.concat([cv.labcodes for cv in self.results], ignore_index=True)  # type: ignore
        else:
            labcodes = None
        if self.results[0].X is not None:
            X = pd.concat([cv.X for cv in self.results], ignore_index=True)  # type: ignore
        else:
            X = None
        return CvResult(
            key=self.results[0].key,
            observed=observed,
            predicted=predicted,
            standard_deviation=sd,
            labcodes=labcodes,
            X=X,
        )

    def get_metric(
        self,
        metric: Union[RegressionMetricsEnum, UQRegressionMetricsEnum],
        combine_folds: bool = True,
    ) -> pd.Series:
        """Calculates a metric for every fold and returns them as pd.Series.

        Args:
            metric (RegressionMetricsEnum): Metrics to calculate.
            combine_folds (bool, optional): If True the data in the split is combined before
                the metric is calculated. In this case only a single number is returned. If False
                the metric is calculated per fold. Defaults to True.

        Returns:
            pd.Series: Object containing the metric value for every fold.
        """
        if self.is_loo or combine_folds:
            return pd.Series(
                self._combine_folds().get_metric(metric=metric), name=metric.name
            )
        return pd.Series(
            [cv.get_metric(metric) for cv in self.results], name=metric.name
        )

    def get_metrics(
        self,
        metrics: Sequence[Union[RegressionMetricsEnum, UQRegressionMetricsEnum]] = [
            RegressionMetricsEnum.MAE,
            RegressionMetricsEnum.MSD,
            RegressionMetricsEnum.R2,
            RegressionMetricsEnum.MAPE,
            RegressionMetricsEnum.PEARSON,
            RegressionMetricsEnum.SPEARMAN,
            RegressionMetricsEnum.FISHER,
        ],
        combine_folds: bool = True,
    ) -> pd.DataFrame:
        """Calculates all metrics provided as list for every fold and returns all as pd.DataFrame.

        Args:
            metrics (Sequence[RegressionMetricsEnum], optional): Metrics to calculate. Defaults to R2, MAE, MSD, R2, MAPE.
            combine_folds (bool, optional): If True the data in the split is combined before
                the metric is calculated. In this case only a single number per metric is returned. If False
                the metric is calculated per fold. Defaults to True.

        Returns:
            pd.DataFrame: Dataframe containing the metric values for all folds.
        """
        return pd.concat([self.get_metric(m, combine_folds) for m in metrics], axis=1)


# the following methods tranform a CvResults object to a CrossValidationValues object
# in which the metrics are stored and not computed on the fly, moreover the field types
# are more backend friendly. It should be used to store CvResults in a backend system
class CrossValidationValues(BaseModel):
    observed: List[float] = Field(description="actual output values")
    predicted: List[float] = Field(description="predicted output values")
    standardDeviation: Optional[List[float]] = Field(
        description="standard deviation of predicted values", default=None
    )
    metrics: Optional[Dict[str, float]] = Field(
        description="metrics per cv fold. Key is the metric type", default=None
    )


def CvResults2CrossValidationValues(
    cv: CvResults,
) -> Dict[str, List[CrossValidationValues]]:
    cvResults = {cv.key: []}
    metrics = cv.get_metrics(combine_folds=False)
    for i, fold in enumerate(cv.results):
        cvResults[cv.key].append(
            CrossValidationValues(
                observed=fold.observed.tolist(),
                predicted=fold.predicted.tolist(),
                standardDeviation=fold.standard_deviation.tolist()
                if fold.standard_deviation is not None
                else None,
                metrics=metrics.loc[i].to_dict() if fold.n_samples > 1 else None,
            )
        )
    return cvResults



from bofire.data_models.strategies.api import (
    AlwaysTrueCondition,
    CombiCondition,
    NumberOfExperimentsCondition,
)
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    AlwaysTrueCondition,
    lambda: {},
)

specs.add_valid(
    NumberOfExperimentsCondition,
    lambda: {"n_experiments": 10},
)

specs.add_valid(
    CombiCondition,
    lambda: {
        "conditions": [
            NumberOfExperimentsCondition(n_experiments=5).dict(),
            AlwaysTrueCondition().dict(),
        ],
        "n_required_conditions": 2,
    },
)



from abc import abstractmethod
from typing import Union

import numpy as np
import pandas as pd
from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel


class Objective(BaseModel):
    """The base class for all objectives"""

    type: str

    @abstractmethod
    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """Abstract method to define the call function for the class Objective

        Args:
            x (np.ndarray): An array of x values

        Returns:
            np.ndarray: The desirability of the passed x values
        """
        pass


# TODO: should this inherit from Objective?
class ConstrainedObjective:
    """This abstract class offers a convenience routine for transforming sigmoid based objectives to botorch output constraints."""


TGt0 = Annotated[float, Field(type=float, gt=0)]
TGe0 = Annotated[float, Field(type=float, ge=0)]
TWeight = Annotated[float, Field(type=float, gt=0, le=1)]



import base64
import io

import numpy as np
import pandas as pd
import torch
from botorch.models.transforms.input import ChainedInputTransform, FilterFeatures

from bofire.data_models.surrogates.api import BotorchSurrogate as DataModel
from bofire.surrogates.surrogate import Surrogate
from bofire.utils.torch_tools import tkwargs


class BotorchSurrogate(Surrogate):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _predict(self, transformed_X: pd.DataFrame):
        # transform to tensor
        X = torch.from_numpy(transformed_X.values).to(**tkwargs)
        with torch.no_grad():
            preds = self.model.posterior(X=X, observation_noise=True).mean.cpu().detach().numpy()  # type: ignore
            stds = np.sqrt(self.model.posterior(X=X, observation_noise=True).variance.cpu().detach().numpy())  # type: ignore
        return preds, stds

    @property
    def is_compatibilized(self) -> bool:
        if self.is_fitted:
            if hasattr(self.model, "input_transform"):
                if self.model.input_transform is not None:  # type: ignore
                    if isinstance(self.model.input_transform, FilterFeatures):  # type: ignore
                        return True
                    if isinstance(self.model.input_transform, ChainedInputTransform):  # type: ignore
                        if "tcompatibilize" in self.model.input_transform.keys():  # type: ignore
                            return True
        return False

    def decompatibilize(self):
        if self.is_fitted:
            if self.is_compatibilized:
                if isinstance(self.model.input_transform, FilterFeatures):  # type: ignore
                    self.model.input_transform = None  # type: ignore
                elif isinstance(self.model.input_transform, ChainedInputTransform):  # type: ignore
                    self.model.input_transform = self.model.input_transform.tf2  # type: ignore
                else:
                    raise ValueError("Undefined input transform structure detected.")

    def _prepare_for_dump(self):
        """Decompatibilize the model before the dump"""
        self.decompatibilize()

    def _dumps(self) -> str:
        """Dumps the actual model to a string via pickle as this is not directly json serializable."""
        # empty internal caches to get smaller dumps
        self.model.prediction_strategy = None
        buffer = io.BytesIO()
        torch.save(self.model, buffer)
        return base64.b64encode(buffer.getvalue()).decode()

    def loads(self, data: str):
        """Loads the actual model from a base64 encoded pickle bytes object and writes it to the `model` attribute."""
        buffer = io.BytesIO(base64.b64decode(data.encode()))
        self.model = torch.load(buffer)



from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
import pandas as pd
from pydantic import PositiveInt

from bofire.data_models.strategies.api import Strategy as DataModel
from bofire.strategies.data_models.candidate import Candidate
from bofire.strategies.data_models.values import InputValue


class Strategy(ABC):
    """Base class for all strategies

    Attributes:
    """

    def __init__(
        self,
        data_model: DataModel,
    ):
        self.domain = data_model.domain
        self.seed = data_model.seed or np.random.default_rng().integers(1000)
        self.rng = np.random.default_rng(self.seed)  # type: ignore
        self._experiments = None
        self._candidates = None

    @classmethod
    def from_spec(cls, data_model: DataModel) -> "Strategy":
        """Used by the mapper to map from data model to functional strategy."""
        return cls(data_model=data_model)

    @property
    def experiments(self) -> pd.DataFrame:
        """Returns the experiments of the strategy.

        Returns:
            pd.DataFrame: Current experiments.
        """
        return self._experiments  # type: ignore

    @property
    def candidates(self) -> pd.DataFrame:
        """Returns the (pending) candidates of the strategy.

        Returns:
            pd.DataFrame: Pending experiments.
        """
        return self._candidates  # type: ignore

    def tell(
        self,
        experiments: pd.DataFrame,
        replace: bool = False,
    ) -> None:
        """This function passes new experimental data to the optimizer

        Args:
            experiments (pd.DataFrame): DataFrame with experimental data
            replace (bool, optional): Boolean to decide if the experimental data should replace the former DataFrame or if the new experiments should be attached. Defaults to False.
        """
        if len(experiments) == 0:
            return
        if replace:
            self.set_experiments(experiments=experiments)
        else:
            self.add_experiments(experiments=experiments)
        # we check here that the experiments do not have completely fixed columns
        cleaned_experiments = (
            self.domain.outputs.preprocess_experiments_all_valid_outputs(
                experiments=experiments
            )
        )
        for feature in self.domain.inputs.get_fixed():
            if (cleaned_experiments[feature.key] == feature.fixed_value()[0]).all():  # type: ignore
                raise ValueError(
                    f"No variance in experiments for fixed feature {feature.key}"
                )
        self._tell()

    def _tell(self) -> None:
        """Method to allow for customized tell functions in addition to self.tell()"""
        pass

    def ask(
        self,
        candidate_count: Optional[PositiveInt] = None,
        add_pending: bool = False,
        raise_validation_error: bool = True,
    ) -> pd.DataFrame:
        """Function to generate new candidates.

        Args:
            candidate_count (PositiveInt, optional): Number of candidates to be generated. If not provided, the number
                of candidates is determined automatically. Defaults to None.
            add_pending (bool, optional): If true the proposed candidates are added to the set of pending experiments. Defaults to False.
            raise_validation_error (bool, optional): If true an error will be raised if candidates violate constraints,
                otherwise only a warning will be displayed. Defaults to True.


        Raises:
            ValueError: if candidate count is smaller than 1
            ValueError: if not enough experiments are available to execute the strategy
            ValueError: if the number of generated candidates does not match the requested number

        Returns:
            pd.DataFrame: DataFrame with candidates (proposed experiments)
        """
        if candidate_count is not None and candidate_count < 1:
            raise ValueError(
                f"Candidate_count has to be at least 1 but got {candidate_count}."
            )
        if not self.has_sufficient_experiments():
            raise ValueError(
                "Not enough experiments available to execute the strategy."
            )

        candidates = self._ask(candidate_count=candidate_count)

        self.domain.validate_candidates(
            candidates=candidates,
            only_inputs=True,
            raise_validation_error=raise_validation_error,
        )

        if candidate_count is not None:
            if len(candidates) != candidate_count:
                raise ValueError(
                    f"expected {candidate_count} candidates, got {len(candidates)}"
                )

        if add_pending:
            self.add_candidates(candidates)

        return candidates

    @abstractmethod
    def has_sufficient_experiments(
        self,
    ) -> bool:
        """Abstract method to check if sufficient experiments are available.

        Returns:
            bool: True if number of passed experiments is sufficient, False otherwise
        """
        pass

    @abstractmethod
    def _ask(
        self,
        candidate_count: Optional[PositiveInt] = None,
    ) -> pd.DataFrame:
        """Abstract method to implement how a strategy generates candidates.

        Args:
            candidate_count (PositiveInt, optional): Number of candidates to be generated. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame with candidates (proposed experiments).
        """
        pass

    def to_candidates(self, candidates: pd.DataFrame) -> List[Candidate]:
        """Transform candiadtes dataframe to a list of `Candidate` objects.
        Args:
            candidates (pd.DataFrame): candidates formatted as dataframe
        Returns:
            List[Candidate]: candidates formatted as list of `Candidate` objects.
        """
        return [
            Candidate(
                inputValues={
                    key: InputValue(value=row[key])
                    for key in self.domain.inputs.get_keys()
                },
            )
            for _, row in candidates.iterrows()
        ]

    def set_candidates(self, candidates: pd.DataFrame):
        """Set pending candidates of the strategy. Overwrites existing ones.

        Args:
            experiments (pd.DataFrame): Dataframe with candidates.
        """
        candidates = self.domain.inputs.validate_experiments(
            candidates[self.domain.inputs.get_keys()], strict=False
        )
        self._candidates = candidates[self.domain.inputs.get_keys()]

    def add_candidates(self, candidates: pd.DataFrame):
        """Add pending candidates to the strategy. Appends to existing ones.

        Args:
            experiments (pd.DataFrame): Dataframe with candidates.
        """
        candidates = self.domain.inputs.validate_experiments(
            candidates[self.domain.inputs.get_keys()], strict=False
        )
        if self.candidates is None:
            self._candidates = candidates[self.domain.inputs.get_keys()]
        else:
            self._candidates = pd.concat(
                (self.candidates, candidates[self.domain.inputs.get_keys()]),
                ignore_index=True,
            )  # type: ignore

    def reset_candidates(self):
        """Resets the pending candidates of the strategy."""
        self._candidates = None

    @property
    def num_candidates(self) -> int:
        """Returns number of (pending) candidates"""
        if self.candidates is None:
            return 0
        return len(self.candidates)

    def set_experiments(self, experiments: pd.DataFrame):
        """Set experiments of the strategy. Overwrites existing ones.

        Args:
            experiments (pd.DataFrame): Dataframe with experiments.
        """
        experiments = self.domain.validate_experiments(experiments)
        self._experiments = experiments

    def add_experiments(self, experiments: pd.DataFrame):
        """Add experiments to the strategy. Appends to existing ones.

        Args:
            experiments (pd.DataFrame): Dataframe with experiments.
        """
        experiments = self.domain.validate_experiments(experiments)
        if self.experiments is None:
            self._experiments = experiments
        else:
            self._experiments = pd.concat(
                (self.experiments, experiments), ignore_index=True
            )

    @property
    def num_experiments(self) -> int:
        """Returns number of experiments"""
        if self.experiments is None:
            return 0
        return len(self.experiments)



from typing import Dict, Type

import bofire.data_models.strategies.api as data_models
from bofire.strategies.doe_strategy import DoEStrategy  # noqa: F401
from bofire.strategies.factorial import FactorialStrategy
from bofire.strategies.predictives.botorch import BotorchStrategy  # noqa: F401
from bofire.strategies.predictives.mobo import MoboStrategy
from bofire.strategies.predictives.predictive import PredictiveStrategy  # noqa: F401
from bofire.strategies.predictives.qehvi import QehviStrategy  # noqa: F401
from bofire.strategies.predictives.qnehvi import QnehviStrategy  # noqa: F401
from bofire.strategies.predictives.qparego import QparegoStrategy  # noqa: F401
from bofire.strategies.predictives.sobo import (  # noqa: F401
    AdditiveSoboStrategy,
    CustomSoboStrategy,
    MultiplicativeSoboStrategy,
    SoboStrategy,
)
from bofire.strategies.random import RandomStrategy  # noqa: F401
from bofire.strategies.samplers.polytope import PolytopeSampler  # noqa: F401
from bofire.strategies.samplers.rejection import RejectionSampler  # noqa: F401
from bofire.strategies.samplers.sampler import SamplerStrategy  # noqa: F401
from bofire.strategies.stepwise.stepwise import StepwiseStrategy
from bofire.strategies.strategy import Strategy  # noqa: F401

STRATEGY_MAP: Dict[Type[data_models.Strategy], Type[Strategy]] = {
    data_models.RandomStrategy: RandomStrategy,
    data_models.SoboStrategy: SoboStrategy,
    data_models.AdditiveSoboStrategy: AdditiveSoboStrategy,
    data_models.MultiplicativeSoboStrategy: MultiplicativeSoboStrategy,
    data_models.CustomSoboStrategy: CustomSoboStrategy,
    data_models.QehviStrategy: QehviStrategy,
    data_models.QnehviStrategy: QnehviStrategy,
    data_models.QparegoStrategy: QparegoStrategy,
    data_models.PolytopeSampler: PolytopeSampler,
    data_models.RejectionSampler: RejectionSampler,
    data_models.DoEStrategy: DoEStrategy,
    data_models.StepwiseStrategy: StepwiseStrategy,
    data_models.FactorialStrategy: FactorialStrategy,
    data_models.MoboStrategy: MoboStrategy,
}


def map(data_model: data_models.Strategy) -> Strategy:
    cls = STRATEGY_MAP[data_model.__class__]
    return cls.from_spec(data_model=data_model)
